import random
import string
from contextlib import closing
from dataclasses import asdict
from typing import cast

import dask.array
import numpy
import pandas
import pandas.testing
import pyarrow
import pytest
import pytest_asyncio
import sqlalchemy.exc
import tifffile
import xarray
from sqlalchemy import text
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.pool import AsyncAdaptedQueuePool, QueuePool, StaticPool

from ..adapters.csv import CSVAdapter
from ..adapters.dataframe import ArrayAdapter
from ..adapters.tiff import TiffAdapter
from ..catalog import in_memory
from ..catalog.adapter import WouldDeleteData
from ..catalog.explain import record_explanations
from ..client import Context, from_context
from ..client.register import register
from ..client.utils import ClientError
from ..client.xarray import write_xarray_dataset
from ..queries import Eq, Key
from ..server.app import build_app, build_app_from_config
from ..server.schemas import Asset, DataSource, Management
from ..storage import SQLStorage, get_storage, parse_storage, sanitize_uri
from ..structures.core import StructureFamily
from ..utils import Conflicts, ensure_specified_sql_driver, ensure_uri
from .utils import sql_table_exists


@pytest_asyncio.fixture
async def a(adapter):
    "Raw adapter, not to be used within an app because it is manually started and stopped."
    await adapter.startup()
    yield adapter
    await adapter.shutdown()


@pytest_asyncio.fixture
async def client(adapter):
    app = build_app(adapter)
    with Context.from_app(app) as context:
        yield from_context(context)


@pytest.mark.asyncio
async def test_nested_node_creation(a):
    await a.create_node(
        key="b",
        metadata={},
        structure_family=StructureFamily.container,
        specs=[],
    )
    b = await a.lookup_adapter(["b"])
    await b.create_node(
        key="c",
        metadata={},
        structure_family=StructureFamily.container,
        specs=[],
    )
    c = await b.lookup_adapter(["c"])
    assert await b.path_segments() == ["b"]
    assert await c.path_segments() == ["b", "c"]
    assert (await a.keys_range(0, 1)) == ["b"]
    assert (await b.keys_range(0, 1)) == ["c"]
    # smoke test
    await a.items_range(0, 1)
    await b.items_range(0, 1)
    await a.shutdown()


@pytest.mark.asyncio
async def test_sorting(a):
    # Generate lists of letters and numbers, randomly shuffled.
    random_state = random.Random(0)
    ordered_letters = list(string.ascii_lowercase[:10])
    shuffled_letters = list(ordered_letters)
    random_state.shuffle(shuffled_letters)
    shuffled_numbers = [0] * 5 + [1] * 5
    random_state.shuffle(shuffled_numbers)
    assert ordered_letters != shuffled_letters
    assert sorted(shuffled_letters) != shuffled_letters

    for letter, number in zip(shuffled_letters, shuffled_numbers):
        await a.create_node(
            key=letter,
            metadata={"letter": letter, "number": number},
            structure_family=StructureFamily.container,
            specs=[],
        )

    # Default sorting is _not_ ordered.
    default_key_order = await a.keys_range(0, 10)
    assert default_key_order != ordered_letters
    # Sorting by ("", -1) gives reversed default order.
    reversed_default_key_order = await a.sort([("", -1)]).keys_range(0, 10)
    assert reversed_default_key_order == list(reversed(default_key_order))

    # Sort by key.
    assert await a.sort([("id", 1)]).keys_range(0, 10) == ordered_letters
    # Test again, with items_range.
    assert [
        k for k, v in await a.sort([("id", 1)]).items_range(0, 10)
    ] == ordered_letters

    # Sort by letter metadata.
    # Use explicit 'metadata.{key}' namespace.
    assert (await a.sort([("metadata.letter", 1)]).keys_range(0, 10)) == ordered_letters

    # Sort by letter metadata.
    # Use implicit '{key}' (more convenient, and necessary for back-compat).
    assert (await a.sort([("letter", 1)]).keys_range(0, 10)) == ordered_letters

    # Sort by number and then by letter.
    # Use explicit 'metadata.{key}' namespace.
    items = await a.sort([("metadata.number", 1), ("metadata.letter", 1)]).items_range(
        0, 10
    )
    numbers = [v.metadata()["number"] for k, v in items]
    letters = [v.metadata()["letter"] for k, v in items]
    keys = [k for k, v in items]
    # Numbers are sorted.
    numbers = sorted(numbers)
    # Within each block of numbers, keys and letters are sorted.
    assert sorted(keys[:5]) == keys[:5] == letters[:5]
    assert sorted(keys[5:]) == keys[5:] == letters[5:]


@pytest.mark.asyncio
async def test_search(a):
    for letter, number in zip(string.ascii_lowercase[:5], range(5)):
        await a.create_node(
            key=letter,
            metadata={"letter": letter, "number": number, "x": {"y": {"z": letter}}},
            structure_family=StructureFamily.container,
            specs=[],
        )
    assert "c" in await a.keys_range(0, 5)
    assert await a.search(Eq("letter", "c")).keys_range(0, 5) == ["c"]
    assert await a.search(Eq("number", 2)).keys_range(0, 5) == ["c"]

    # Looking up "d" inside search results should find nothing when
    # "d" is filtered out by a search query first.
    assert await a.lookup_adapter(["d"]) is not None
    with pytest.raises(KeyError):
        await a.search(Eq("letter", "c")).lookup_adapter(["d"])

    # Search on nested key.
    assert await a.search(Eq("x.y.z", "c")).keys_range(0, 5) == ["c"]

    # Created nested nodes and search on them.
    d = await a.lookup_adapter(["d"])
    for letter, number in zip(string.ascii_lowercase[:5], range(10, 15)):
        await d.create_node(
            key=letter,
            metadata={"letter": letter, "number": number},
            structure_family=StructureFamily.container,
            specs=[],
        )
    assert await d.search(Eq("letter", "c")).keys_range(0, 5) == ["c"]
    assert await d.search(Eq("number", 12)).keys_range(0, 5) == ["c"]


@pytest.mark.asyncio
async def test_metadata_index_is_used(example_data_adapter):
    a = example_data_adapter  # for succinctness below
    # Check that an index (specifically the 'top_level_metadata' index) is used
    # by inspecting the content of an 'EXPLAIN ...' query. The exact content
    # is intended for humans and is not an API, but we can coarsely check
    # that the index of interest is mentioned.
    await a.startup()
    with record_explanations() as e:
        results = await a.search(Key("number_as_string") == "3").keys_range(0, 5)
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("number") == 3).keys_range(0, 5)
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("bool") == False).keys_range(0, 5)  # noqa: #712
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("nested.number_as_string") == "3").keys_range(0, 5)
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("nested.number") == 3).keys_range(0, 5)
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("nested.bool") == False).keys_range(  # noqa: #712
            0, 5
        )
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    await a.shutdown()


@pytest.mark.asyncio
async def test_write_array_external(a, tmpdir):
    arr = numpy.ones((5, 3))
    filepath = str(tmpdir / "file.tiff")
    data_uri = ensure_uri(filepath)
    tifffile.imwrite(filepath, arr)
    ad = TiffAdapter(data_uri)
    structure = asdict(ad.structure())
    await a.create_node(
        key="x",
        structure_family="array",
        metadata={},
        data_sources=[
            DataSource(
                structure_family="array",
                mimetype="image/tiff",
                structure=structure,
                parameters={},
                management="external",
                assets=[
                    Asset(
                        parameter="data_uri",
                        num=None,
                        data_uri=str(data_uri),
                        is_directory=False,
                    )
                ],
            )
        ],
    )
    x = await a.lookup_adapter(["x"])
    assert numpy.array_equal(await x.read(), arr)


@pytest.mark.asyncio
async def test_write_dataframe_external_direct(a, tmpdir):
    df = pandas.DataFrame(numpy.ones((5, 3)), columns=list("abc"))
    filepath = str(tmpdir / "file.csv")
    data_uri = ensure_uri(filepath)
    df.to_csv(filepath, index=False)
    dfa = CSVAdapter.from_uris(data_uri)
    structure = asdict(dfa.structure())
    await a.create_node(
        key="x",
        structure_family=StructureFamily.table,
        metadata={},
        data_sources=[
            DataSource(
                structure_family="table",
                mimetype="text/csv",
                structure=structure,
                parameters={},
                management="external",
                assets=[
                    Asset(
                        parameter="data_uris",
                        num=0,
                        data_uri=data_uri,
                        is_directory=False,
                    )
                ],
            )
        ],
    )
    x = await a.lookup_adapter(["x"])
    pandas.testing.assert_frame_equal(await x.read(), df)


@pytest.mark.asyncio
async def test_write_array_internal_direct(a, tmpdir):
    from ..media_type_registration import default_deserialization_registry

    arr = numpy.ones((5, 3))
    ad = ArrayAdapter.from_array(arr)
    structure = ad.structure()
    await a.create_node(
        key="x",
        structure_family="array",
        metadata={},
        data_sources=[
            DataSource(
                structure_family="array",
                structure=structure,
                management="writable",
            )
        ],
    )
    x = await a.lookup_adapter(["x"])

    media_type = "application/octet-stream"
    body = arr.tobytes()
    deserializer = default_deserialization_registry.dispatch("array", media_type)
    await x.write(media_type, deserializer, x, body)

    val = await x.read()
    assert numpy.array_equal(val, arr)


def test_write_array_internal_via_client(client):
    expected = numpy.array([1, 3, 7])
    x = client.write_array(expected)
    actual = x.read()
    assert numpy.array_equal(actual, expected)

    y = client.write_array(dask.array.from_array(expected, chunks=((1, 1, 1),)))
    actual = y.read()
    assert numpy.array_equal(actual, expected)


def test_write_dataframe_internal_via_client(client):
    expected = pandas.DataFrame(numpy.ones((5, 3)), columns=list("abc"))
    x = client.write_dataframe(expected)
    actual = x.read()
    pandas.testing.assert_frame_equal(actual, expected)

    # y = client.write_array(dask.array.from_array(expected, chunks=((1, 1, 1),)))
    # actual = y.read()
    # assert numpy.array_equal(actual, expected)
    # pandas.testing.assert_frame_equal(actual, expected)


def test_write_xarray_dataset(client):
    ds = xarray.Dataset(
        {"temp": (["time"], numpy.array([101, 102, 103]))},
        coords={"time": (["time"], numpy.array([1, 2, 3]))},
    )
    dsc = write_xarray_dataset(client, ds, key="test_xarray_dataset")
    assert set(dsc) == {"temp", "time"}
    # smoke test
    dsc["temp"][:]
    dsc["time"][:]
    dsc.read()


@pytest.mark.asyncio
async def test_delete_catalog_tree(tmpdir):
    # Do not use client fixture here.
    # The Context must be opened inside the test or we run into
    # event loop crossing issues with the Postgres test.
    tree = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        a = client.create_container("a")
        b = a.create_container("b")
        b.write_array([1, 2, 3])
        b.write_array([4, 5, 6])
        c = b.create_container("c")
        d = c.create_container("d")
        d.write_array([7, 8, 9])

        nodes_before_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_before_delete) == 7 + 1  # +1 for the root node
        data_sources_before_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_before_delete) == 3
        assets_before_delete = (
            await tree.context.execute("SELECT * from assets")
        ).all()
        assert len(assets_before_delete) == 3

        with pytest.raises(Conflicts, match="Cannot delete a node that is not empty."):
            await tree.delete()

        with pytest.raises(WouldDeleteData):
            await tree.delete(recursive=True)  # external_only=True by default
        with pytest.raises(WouldDeleteData):
            await tree.delete(recursive=True, external_only=True)
        await tree.delete(recursive=True, external_only=False)

        nodes_after_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_after_delete) == 0 + 1  # the root node that should remain
        data_sources_after_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_after_delete) == 0
        assets_after_delete = (await tree.context.execute("SELECT * from assets")).all()
        assert len(assets_after_delete) == 0


@pytest.mark.asyncio
async def test_delete_contents(tmpdir):
    # Do not use client fixture here.
    # The Context must be opened inside the test or we run into
    # event loop crossing issues with the Postgres test.
    tree = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        # a has children b1 and b2, which each contain arrays
        a = client.create_container("a")
        b1 = a.create_container("b1")
        b1.write_array([1, 2, 3], key="test_1")
        b1.write_array([4, 5, 6], key="test_2")
        b1.write_array([7, 8, 9], key="test_3")
        b2 = a.create_container("b2")
        b2.write_array([10, 11, 12], key="test_4")
        b2.write_array([13, 14, 15], key="test_5")
        a.create_container("b3")  # empty container

        assert set(client) == {"a"}
        assert set(client["a"]) == {"b1", "b2", "b3"}
        assert set(client["a"]["b1"]) == {"test_1", "test_2", "test_3"}
        assert set(client["a"]["b2"]) == {"test_4", "test_5"}

        # Check the database state before deletion
        nodes_before_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_before_delete) == 9 + 1
        data_sources_before_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_before_delete) == 5
        assets_before_delete = (
            await tree.context.execute("SELECT * from assets")
        ).all()
        assert len(assets_before_delete) == 5

        # Trying to delete a non-empty node without recursive=True should raise
        with pytest.raises(
            ClientError, match="Cannot delete a node that is not empty."
        ):
            client["a"].delete_contents(["b1"], recursive=False, external_only=True)

        # Trying to delete internal data with external_only=True should raise
        with pytest.raises(
            ClientError, match="Some items in this tree are internally managed."
        ):
            client["a"].delete_contents(["b1"], recursive=True, external_only=True)

        # Delete arrays from b1 (as a scalar and as a list), and then b1 itself
        b1.delete_contents("test_1", external_only=False)
        assert set(client["a"]["b1"].keys()) == {"test_2", "test_3"}
        b1.delete_contents(["test_2", "test_3"], external_only=False)
        assert set(client["a"]["b1"].keys()) == set()
        client["a"].delete_contents(["b1"], recursive=False, external_only=True)
        assert set(client["a"]) == {"b2", "b3"}

        # Delete all contents of a, including the non-empty b2 and the empty b3
        client["a"].delete_contents(external_only=False, recursive=True)
        assert set(client["a"]) == set()

        # Check the database state; only a and the root node should remain.
        nodes_after_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_after_delete) == 1 + 1
        data_sources_after_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_after_delete) == 0
        assets_after_delete = (await tree.context.execute("SELECT * from assets")).all()
        assert len(assets_after_delete) == 0


@pytest.mark.asyncio
async def test_delete_with_external_nodes(tmpdir):
    # Do not use client fixture here.
    # The Context must be opened inside the test or we run into
    # event loop crossing issues with the Postgres test.
    (tmpdir / "readable").mkdir()
    (tmpdir / "writable").mkdir()
    tree = in_memory(
        readable_storage=[str(tmpdir / "readable")],
        writable_storage={"filesystem": str(tmpdir / "writable")},
    )

    # Create some external data to register
    for i in range(1, 5):
        with open(tmpdir / "readable" / f"test_{i}.csv", "w") as file:
            file.write(
                """a, b, c
                    1, 2, 3
                    4, 5, 6
                """
            )
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        # a has children b1 and b2, which each contain arrays
        a = client.create_container("a")
        b1 = a.create_container("b1")
        await register(b1, tmpdir / "readable" / "test_1.csv")
        await register(b1, tmpdir / "readable" / "test_2.csv")
        b2 = a.create_container("b2")
        await register(b2, tmpdir / "readable" / "test_3.csv")
        await register(b2, tmpdir / "readable" / "test_4.csv")

        assert list(client) == ["a"]
        assert list(client["a"]) == ["b1", "b2"]
        assert list(client["a"]["b1"]) == ["test_1", "test_2"]
        assert list(client["a"]["b2"]) == ["test_3", "test_4"]

        nodes_before_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_before_delete) == 7 + 1  # +1 for the root node
        data_sources_before_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_before_delete) == 4
        assets_before_delete = (
            await tree.context.execute("SELECT * from assets")
        ).all()
        assert len(assets_before_delete) == 4

        # Delete all children of b1, and b1 itself.
        client["a"].delete_contents("b1", recursive=True)

        assert list(client) == ["a"]
        assert list(client["a"]) == ["b2"]
        assert list(client["a"]["b2"]) == ["test_3", "test_4"]  # not affected
        nodes_after_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_after_delete) == 4 + 1  # +1 for the root node
        data_sources_after_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_after_delete) == 2
        assets_after_delete = (await tree.context.execute("SELECT * from assets")).all()
        assert len(assets_after_delete) == 2


@pytest.mark.asyncio
async def test_delete_sql_assets(sql_storage_uri):
    # Do not use client fixture here.
    # The Context must be opened inside the test or we run into
    # event loop crossing issues with the Postgres test.

    tree = in_memory(writable_storage={"sql": sql_storage_uri})
    storage = cast(SQLStorage, get_storage(parse_storage(sql_storage_uri).uri))

    # Create some tables to write
    table_1 = pyarrow.Table.from_pydict({"a": [1, 2, 3], "b": [4.0, 5.0, 6.0]})
    table_2 = pyarrow.Table.from_pydict({"c": [4, 5, 6], "d": ["7", "8", "9"]})

    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        # a has children b1 and b2, which each contain arrays
        a = client.create_container("a")
        b1 = a.create_container("b1")
        t1 = b1.create_appendable_table(schema=table_1.schema, key="table_1")
        t1.append_partition(table_1, 0)
        t1.append_partition(table_1, 0)
        t2 = b1.create_appendable_table(schema=table_2.schema, key="table_2")
        t2.append_partition(table_2, 0)
        assert t1.read() is not None
        assert t2.read() is not None

        # Check the SQL storage directly
        t1_table_name = t1.data_sources()[0].parameters["table_name"]
        t1_dataset_id = t1.data_sources()[0].parameters["dataset_id"]
        t2_table_name = t2.data_sources()[0].parameters["table_name"]
        t2_dataset_id = t2.data_sources()[0].parameters["dataset_id"]
        with closing(storage.connect()) as conn:
            assert sql_table_exists(conn, storage.dialect, t1_table_name)
            assert sql_table_exists(conn, storage.dialect, t2_table_name)
            with conn.cursor() as cursor:
                cursor.execute(
                    f'SELECT COUNT(*) FROM "{t1_table_name}" '
                    f"WHERE _dataset_id = {t1_dataset_id:d};",
                )
                assert cursor.fetchone()[0] == 6
                cursor.execute(
                    f'SELECT COUNT(*) FROM "{t2_table_name}" '
                    f"WHERE _dataset_id = {t2_dataset_id:d};",
                )
                assert cursor.fetchone()[0] == 3

        # Add another table to b2 -- a copy of table_1 with the same schema
        b2 = a.create_container("b2")
        t1c = b2.create_appendable_table(schema=table_1.schema, key="table_1_copy")
        t1c.append_partition(table_1, 0)
        assert t1c.read() is not None

        # Check the catalog state before deletion
        assert list(client) == ["a"]
        assert list(client["a"]) == ["b1", "b2"]
        assert list(client["a"]["b1"]) == ["table_1", "table_2"]
        assert list(client["a"]["b2"]) == ["table_1_copy"]

        # Check the number of nodes, data sources, and assets
        nodes_before_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_before_delete) == 6 + 1  # +1 for the root node
        data_sources_before_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_before_delete) == 3
        assets_before_delete = (
            await tree.context.execute("SELECT * from assets")
        ).all()
        assert len(assets_before_delete) == 1  # single sql asset

        # Check the SQL storage directly
        t1c_table_name = t1c.data_sources()[0].parameters["table_name"]
        t1c_dataset_id = t1c.data_sources()[0].parameters["dataset_id"]
        assert t1c_table_name == t1_table_name
        assert t1c_dataset_id != t1_dataset_id
        with closing(storage.connect()) as conn:
            with conn.cursor() as cursor:
                cursor.execute(
                    f'SELECT COUNT(*) FROM "{t1_table_name}";',
                )
                assert cursor.fetchone()[0] == 9
                cursor.execute(
                    f'SELECT COUNT(*) FROM "{t2_table_name}";',
                )
                assert cursor.fetchone()[0] == 3
                cursor.execute(
                    f'SELECT COUNT(*) FROM "{t1c_table_name}" '
                    f"WHERE _dataset_id = {t1c_dataset_id:d};",
                )
                assert cursor.fetchone()[0] == 3

        # Delete all children of b1 (tables t1 and t2), but not b1 itself.
        client["a"]["b1"].delete_contents(
            client["a"]["b1"].keys(), recursive=True, external_only=False
        )
        with closing(storage.connect()) as conn:
            with conn.cursor() as cursor:
                assert sql_table_exists(conn, storage.dialect, t1_table_name)
                cursor.execute(
                    f'SELECT COUNT(*) FROM "{t1_table_name}";',
                )
                assert cursor.fetchone()[0] == 3  # 6 rows deleted
                # Entire t2 deleted
                assert not sql_table_exists(conn, storage.dialect, t2_table_name)

        assert list(client) == ["a"]
        assert list(client["a"]) == ["b1", "b2"]
        assert (
            list(client["a"]["b1"]) == []
        )  # children deleted (2 nodes, 2 data sources, 0 assets)
        assert list(client["a"]["b2"]) == ["table_1_copy"]  # not affected
        nodes_after_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_after_delete) == 4 + 1  # +1 for the root node
        data_sources_after_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_after_delete) == 1
        assets_after_delete = (await tree.context.execute("SELECT * from assets")).all()
        assert len(assets_after_delete) == 1

        # Close and dispose the SQL storage
        storage.dispose()


@pytest.mark.asyncio
async def test_delete_external_asset_registered_twice(tmpdir):
    # Do not use client fixture here.
    # The Context must be opened inside the test or we run into
    # event loop crossing issues with the Postgres test.
    tree = in_memory(readable_storage=[str(tmpdir)])
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        for i in range(1, 4):
            with open(tmpdir / f"test_{i}.csv", "w") as file:
                file.write(
                    """a, b, c
                        1, 2, 3
                        4, 5, 6
                    """
                )
        # a has children b1 and b2, which each contain arrays
        a = client.create_container("a")
        b1 = a.create_container("b1")
        await register(b1, tmpdir / "test_1.csv")
        await register(b1, tmpdir / "test_2.csv")
        b2 = a.create_container("b2")
        await register(b2, tmpdir / "test_1.csv")
        await register(b2, tmpdir / "test_3.csv")

        # test_1.csv is registered in both b1 and b2
        assert client["a"]["b1"]["test_1"].read() is not None
        assert client["a"]["b2"]["test_1"].read() is not None

        data_sources_before_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_before_delete) == 4
        assets_after_delete = (await tree.context.execute("SELECT * from assets")).all()
        assert len(assets_after_delete) == 3  # shared by two data sources

        a.delete_contents("b2", recursive=True)

        data_sources_after_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_after_delete) == 2
        assets_after_delete = (await tree.context.execute("SELECT * from assets")).all()
        assert len(assets_after_delete) == 2

        # The asset in b1 should still be accessible.
        client["a"]["b1"]["test_1"].read()

        a.delete_contents("b1", recursive=True)
        data_sources_after_second_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_after_second_delete) == 0
        assets_after_second_delete = (
            await tree.context.execute("SELECT * from assets")
        ).all()
        assert len(assets_after_second_delete) == 0


@pytest.mark.parametrize(
    "assets",
    [
        [
            Asset(
                data_uri="file://localhost/test1",
                is_directory=False,
                parameter="filepath",
                num=None,
            ),
            Asset(
                data_uri="file://localhost/test2",
                is_directory=False,
                parameter="filepath",
                num=1,
            ),
        ],
        [
            Asset(
                data_uri="file://localhost/test1",
                is_directory=False,
                parameter="filepath",
                num=1,
            ),
            Asset(
                data_uri="file://localhost/test2",
                is_directory=False,
                parameter="filepath",
                num=None,
            ),
        ],
        [
            Asset(
                data_uri="file://localhost/test1",
                is_directory=False,
                parameter="filepath",
                num=None,
            ),
            Asset(
                data_uri="file://localhost/test2",
                is_directory=False,
                parameter="filepath",
                num=None,
            ),
        ],
        [
            Asset(
                data_uri="file://localhost/test1",
                is_directory=False,
                parameter="filepath",
                num=1,
            ),
            Asset(
                data_uri="file://localhost/test2",
                is_directory=False,
                parameter="filepath",
                num=1,
            ),
        ],
    ],
    ids=[
        "null-then-int",
        "int-then-null",
        "duplicate-null",
        "duplicate-int",
    ],
)
@pytest.mark.asyncio
async def test_constraints_on_parameter_and_num(a, assets):
    "Test constraints enforced by database on 'parameter' and 'num'."
    arr_adapter = ArrayAdapter.from_array([1, 2, 3])
    with pytest.raises(
        (
            sqlalchemy.exc.IntegrityError,  # SQLite
            sqlalchemy.exc.DBAPIError,  # PostgreSQL
        )
    ):
        await a.create_node(
            key="test",
            structure_family=arr_adapter.structure_family,
            metadata=dict(arr_adapter.metadata()),
            specs=arr_adapter.specs,
            data_sources=[
                DataSource(
                    structure_family=arr_adapter.structure_family,
                    mimetype="text/csv",
                    structure=arr_adapter.structure(),
                    parameters={},
                    management=Management.external,
                    assets=assets,
                )
            ],
        )


@pytest.mark.asyncio
async def test_init_db_logging(sqlite_or_postgres_uri, tmpdir, caplog):
    config = {
        "database": {
            "uri": "sqlite://",  # in-memory
        },
        "trees": [
            {
                "tree": "catalog",
                "path": "/",
                "args": {
                    "uri": sqlite_or_postgres_uri,
                    "writable_storage": str(tmpdir / "data"),
                    "init_if_not_exists": True,
                },
            },
        ],
    }
    # Issue 721 notes that the logging of the subprocess that creates
    # a database logs normal things to error. This test looks at the log
    # and fails if an error log happens. This could catch anything that is
    # an error during the app build.
    import logging

    with caplog.at_level(logging.INFO):
        app = build_app_from_config(config)
        for record in caplog.records:
            assert record.levelname != "ERROR", f"Error found creating app {record.msg}"
        assert app


@pytest.mark.parametrize(
    "exact_count_limit, expected_lower_bound", [(None, 10), (5, 6), (-1, 10)]
)
@pytest.mark.asyncio
async def test_container_length(
    sqlite_or_postgres_uri, exact_count_limit, expected_lower_bound
):
    config = {
        "trees": [
            {
                "tree": "catalog",
                "path": "/",
                "args": {
                    "uri": sqlite_or_postgres_uri,
                    "init_if_not_exists": True,
                },
            },
        ],
    }
    if exact_count_limit is not None:
        config["exact_count_limit"] = exact_count_limit

    app = build_app_from_config(config)

    # Turn off autovacuum in Postgres (just in case)
    # Create a separate engine to avoid interfeing with the running loop
    if sqlite_or_postgres_uri.startswith("postgresql"):
        engine = create_async_engine(
            ensure_specified_sql_driver(sqlite_or_postgres_uri)
        )
        async with engine.begin() as conn:
            await conn.execute(
                text(
                    """
                    ALTER TABLE nodes
                    SET (autovacuum_enabled = false,
                        autovacuum_analyze_threshold = 0);
                """
                )
            )

    with Context.from_app(app) as context:
        client = from_context(context)

        # Create a container with some nested nodes
        a = client.create_container("a")
        for i in range(10):
            b = a.create_container(key=f"node_{i}")
            b.create_container(key=f"subnode_{i}")

        # Before analyzing the table, the length should be thresholded
        len_from_metadata = client["a"].item["attributes"]["structure"]["count"]
        assert len_from_metadata == expected_lower_bound

        # Analyze the table to get update pg_statistics
        if sqlite_or_postgres_uri.startswith("postgresql"):
            async with engine.connect() as conn:
                conn = await conn.execution_options(isolation_level="AUTOCOMMIT")
                await conn.execute(text("VACUUM ANALYZE nodes;"))
            await engine.dispose()

        # After analyzing, the length should be updated (at least be approximate)
        len_from_metadata = client["a"].item["attributes"]["structure"]["count"]
        assert len_from_metadata <= 10

        # len() returns the exact count
        assert len(client["a"]) == 10


@pytest.mark.parametrize(
    "desired, expected",
    [((None, None, None, None), (5, 5, 10, 10)), ((7, 11, 13, 17), (7, 11, 13, 17))],
)
def test_pooling_config(sqlite_or_postgres_uri, sql_storage_uri, desired, expected):
    config = {
        "trees": [
            {
                "tree": "catalog",
                "path": "/",
                "args": {
                    "uri": sqlite_or_postgres_uri,
                    "writable_storage": sql_storage_uri,
                    "init_if_not_exists": True,
                },
            },
        ],
        "catalog_pool_size": desired[0],
        "storage_pool_size": desired[1],
        "catalog_max_overflow": desired[2],
        "storage_max_overflow": desired[3],
    }

    app = build_app_from_config(config)

    # Check the catalog pool
    catalog_pool = app.state.root_tree.context.engine.pool
    assert isinstance(catalog_pool, AsyncAdaptedQueuePool)
    assert catalog_pool.size() == expected[0]
    assert catalog_pool._max_overflow == expected[2]

    # Check the storage pool
    storage = get_storage(ensure_uri(sanitize_uri(sql_storage_uri)[0]))
    storage: SQLStorage = cast(SQLStorage, storage)

    if sql_storage_uri.startswith("duckdb"):
        # DuckDB does not support pooling
        assert isinstance(storage._connection_pool, StaticPool)
        assert storage.pool_size == 1
        assert storage.max_overflow == 0
    else:
        assert isinstance(storage._connection_pool, QueuePool)
        assert storage.pool_size == expected[1]
        assert storage.max_overflow == expected[3]
        assert storage._connection_pool.size() == expected[1]
        assert storage._connection_pool._max_overflow == expected[3]

    storage.dispose()



from pathlib import Path

import pytest

from ..utils import (
    CachingMap,
    DictView,
    ListView,
    OneShotCachedMap,
    ensure_specified_sql_driver,
    parse_mimetype,
    parse_time_string,
    sanitize_uri,
    walk,
)


def test_ensure_specified_sql_driver():
    # Postgres
    # Default driver is added if missing.
    assert (
        ensure_specified_sql_driver(
            "postgresql://user:password@localhost:5432/database"
        )
        == "postgresql+asyncpg://user:password@localhost:5432/database"
    )
    # Default driver passes through if specified.
    assert (
        ensure_specified_sql_driver(
            "postgresql+asyncpg://user:password@localhost:5432/database"
        )
        == "postgresql+asyncpg://user:password@localhost:5432/database"
    )
    # Do not override user-provided.
    assert (
        ensure_specified_sql_driver(
            "postgresql+custom://user:password@localhost:5432/database"
        )
        == "postgresql+custom://user:password@localhost:5432/database"
    )

    # SQLite
    # Default driver is added if missing.
    assert (
        ensure_specified_sql_driver("sqlite:////test.db")
        == "sqlite+aiosqlite:////test.db"
    )
    # Default driver passes through if specified.
    assert (
        ensure_specified_sql_driver("sqlite+aiosqlite:////test.db")
        == "sqlite+aiosqlite:////test.db"
    )
    # Do not override user-provided.
    assert (
        ensure_specified_sql_driver("sqlite+custom:////test.db")
        == "sqlite+custom:////test.db"
    )
    # Handle SQLite :memory: URIs
    assert (
        ensure_specified_sql_driver("sqlite+aiosqlite://:memory:")
        == "sqlite+aiosqlite://:memory:"
    )
    assert (
        ensure_specified_sql_driver("sqlite://:memory:")
        == "sqlite+aiosqlite://:memory:"
    )
    # Handle SQLite relative URIs
    assert (
        ensure_specified_sql_driver("sqlite+aiosqlite:///test.db")
        == "sqlite+aiosqlite:///test.db"
    )
    assert (
        ensure_specified_sql_driver("sqlite:///test.db")
        == "sqlite+aiosqlite:///test.db"
    )
    # Filepaths are implicitly SQLite databases.
    # Relative path
    assert ensure_specified_sql_driver("test.db") == "sqlite+aiosqlite:///test.db"
    # Path object
    assert ensure_specified_sql_driver(Path("test.db")) == "sqlite+aiosqlite:///test.db"
    # Relative path anchored to .
    assert ensure_specified_sql_driver("./test.db") == "sqlite+aiosqlite:///test.db"
    # Absolute path
    assert (
        ensure_specified_sql_driver(Path("/tmp/test.db"))
        == f"sqlite+aiosqlite:///{Path('/tmp/test.db')}"
    )


@pytest.mark.parametrize(
    "string_input,expected",
    [
        ("3s", 3),
        ("7m", 7 * 60),
        ("5h", 5 * 60 * 60),
        ("1d", 1 * 24 * 60 * 60),
        ("2y", 2 * 365 * 24 * 60 * 60),
    ],
)
def test_parse_time_string_valid(string_input, expected):
    assert parse_time_string(string_input) == expected


@pytest.mark.parametrize(
    "string_input",
    [
        "3z",  # unrecognized units
        "3M",  # unrecognized units
        "-3m",  # invalid character '-'
        "3 m",  # invalid character '-'
    ],
)
def test_parse_time_string_invalid(string_input):
    with pytest.raises(ValueError):
        parse_time_string(string_input)


def test_listview_repr():
    lv = ListView([1, 2, 3])
    assert repr(lv) == "ListView([1, 2, 3])"


def test_dictview_repr():
    dv = DictView({"a": 1, "b": 2})
    assert repr(dv) == "DictView({'a': 1, 'b': 2})"


def test_listview_repr_pretty(monkeypatch):
    lv = ListView([1, 2, 3])
    called = {}

    class DummyP:
        "Dummy pretty printer to capture the text output."

        def text(self, s):
            called["text"] = s

    # Should use pformat on a list
    lv._repr_pretty_(DummyP(), cycle=False)
    assert called["text"] == "[1, 2, 3]"

    # Should convert to list if not a list
    lv2 = ListView((4, 5, 6))
    lv2._internal_list = (4, 5, 6)  # forcibly set to tuple
    called.clear()
    lv2._repr_pretty_(DummyP(), cycle=False)
    assert called["text"] == "[4, 5, 6]"


def test_dictview_repr_pretty(monkeypatch):
    dv = DictView({"a": 1, "b": 2})
    called = {}

    class DummyP:
        "Dummy pretty printer to capture the text output."

        def text(self, s):
            called["text"] = s

    # Should use pformat on a dict
    dv._repr_pretty_(DummyP(), cycle=False)
    assert called["text"] == "{'a': 1, 'b': 2}"

    # Should convert to dict if not a dict
    dv2 = DictView([("x", 10), ("y", 20)])
    dv2._internal_dict = [("x", 10), ("y", 20)]  # forcibly set to list of tuples
    called.clear()
    dv2._repr_pretty_(DummyP(), cycle=False)
    # The order of keys in dict may not be guaranteed, so check both possibilities
    assert called["text"] in ("{'x': 10, 'y': 20}", "{'y': 20, 'x': 10}")


def test_oneshotcachedmap_repr_lazy_and_evaluated():
    # Value factories
    def factory1() -> int:
        return 42

    def factory2() -> str:
        return "foo"

    # All values are lazy initially
    m = OneShotCachedMap(a=factory1, b=factory2)
    r = repr(m)
    assert "<OneShotCachedMap" in r
    assert "'a': <lazy>" in r
    assert "'b': <lazy>" in r

    # Access one value to trigger evaluation
    assert m["a"] == 42
    r2 = repr(m)
    assert "'a': 42" in r2
    assert "'b': <lazy>" in r2

    # Access both
    assert m["b"] == "foo"
    r3 = repr(m)
    assert "'a': 42" in r3
    assert "'b': 'foo'" in r3


def test_cachingmap_repr_lazy_and_evaluated():
    # Value factories
    def factory1() -> int:
        return 123

    def factory2() -> str:
        return "bar"

    mapping = {"x": factory1, "y": factory2}
    cache = {}

    m = CachingMap(mapping.copy(), cache)
    # Initially, nothing is cached, so repr should show <lazy>
    r = repr(m)
    assert "<CachingMap" in r
    assert "'x': <lazy>" in r
    assert "'y': <lazy>" in r

    # Access one value to trigger evaluation and caching
    assert m["x"] == 123
    r2 = repr(m)
    assert "'x': 123" in r2
    assert "'y': <lazy>" in r2

    # Access both
    assert m["y"] == "bar"
    r3 = repr(m)
    assert "'x': 123" in r3
    assert "'y': 'bar'" in r3

    # If cache is None, all should be <lazy>
    m2 = CachingMap(mapping.copy(), None)
    r4 = repr(m2)
    assert "'x': <lazy>" in r4
    assert "'y': <lazy>" in r4


@pytest.mark.parametrize(
    "mapping, expected",
    [
        (
            {
                "A": {
                    "dog": {},
                    "cat": {},
                    "monkey": {},
                },
                "B": {
                    "snake": {},
                    "bear": {},
                    "wolf": {},
                },
            },
            [
                ["A"],
                ["A", "dog"],
                ["A", "cat"],
                ["A", "monkey"],
                ["B"],
                ["B", "snake"],
                ["B", "bear"],
                ["B", "wolf"],
            ],
        ),
        ({"root": 42}, [["root"]]),
        ({"x": {"y": {"z": 1}}, "a": 2}, [["x"], ["x", "y"], ["x", "y", "z"], ["a"]]),
        ({}, []),
        ({"foo": object()}, [["foo"]]),
    ],
    ids=["nested_dict", "leaf_value", "nested_mixed", "empty_dict", "non_dict_leaf"],
)
def test_walk(mapping, expected):
    result = list(walk(mapping))
    assert result == expected


@pytest.mark.parametrize(
    "uri, expected_clean_uri, expected_username, expected_password",
    [
        # URI with username and password
        (
            "postgresql://user:pass@localhost:5432/db",
            "postgresql://localhost:5432/db",
            "user",
            "pass",
        ),
        # URI with only username
        (
            "postgresql://user@localhost:5432/db",
            "postgresql://localhost:5432/db",
            "user",
            None,
        ),
        # URI with no username/password
        (
            "postgresql://localhost:5432/db",
            "postgresql://localhost:5432/db",
            None,
            None,
        ),
        # URI with username and password, no port
        (
            "sqlite://user:pass@localhost/db",
            "sqlite://localhost/db",
            "user",
            "pass",
        ),
        # URI with username only, no port
        (
            "sqlite://user@localhost/db",
            "sqlite://localhost/db",
            "user",
            None,
        ),
        # URI with username and password, with query and fragment
        (
            "postgresql://user:pass@localhost:5432/db?foo=bar#frag",
            "postgresql://localhost:5432/db?foo=bar#frag",
            "user",
            "pass",
        ),
        # URI with username only, with query and fragment
        (
            "postgresql://user@localhost:5432/db?foo=bar#frag",
            "postgresql://localhost:5432/db?foo=bar#frag",
            "user",
            None,
        ),
        # URI with no netloc (should not fail)
        (
            "sqlite:///db.sqlite",
            "sqlite:///db.sqlite",
            None,
            None,
        ),
    ],
)
def test_sanitize_uri(uri, expected_clean_uri, expected_username, expected_password):
    clean_uri, username, password = sanitize_uri(uri)
    assert clean_uri == expected_clean_uri
    assert username == expected_username
    assert password == expected_password


@pytest.mark.parametrize(
    "mimetype, expected",
    [
        ("text/csv", ("text/csv", {})),
        ("text/csv;header=absent", ("text/csv", {"header": "absent"})),
        (
            "text/csv;header=absent; charset=utf-8",
            ("text/csv", {"header": "absent", "charset": "utf-8"}),
        ),
        (
            "text/csv; header=absent; charset=utf-8",
            ("text/csv", {"header": "absent", "charset": "utf-8"}),
        ),
    ],
)
def test_parse_valid_mimetype(mimetype, expected):
    assert parse_mimetype(mimetype) == expected


def test_parse_invalid_mimetype():
    with pytest.raises(ValueError):
        # Parameter does not have form 'key=value'
        assert parse_mimetype("text/csv;oops")



import re
from io import StringIO
from pathlib import Path

import pandas as pd
import pytest

from ..client import Context, from_context
from ..client.register import register
from ..examples.xdi import data
from ..server.app import build_app_from_config


def load_data(text: str) -> tuple[list[str], pd.DataFrame]:
    # This regular expression matches and removes a block of comment lines
    #
    # Pattern breakdown:
    #   - #\s*\/+\r?\n
    #       Matches a line starting with '#', optional whitespace, one or more '/',
    #       then a newline(unix or windows).
    #       Example #  //////// newline
    #   - (?:#\s+.*\r?\n)*
    #       Matches zero or more lines starting with '#', at least one space, then any text,
    #       then a newline(unix or windows).
    #       Example #    Comment here newline
    #   - #\s*-+\r?\n
    #       Matches a line starting with '#', optional whitespace, one or more '-',
    #       then a newline(unix or windows).
    #       Example #      -----------newline
    # Example matched block:
    #   # ////////
    #   # Comment Here
    #   # More Comments
    #   # -------
    text = re.sub(r"#\s*\/+\r?\n(?:#\s+.*\r?\n)*#\s*-+\r?\n", "", text)
    lines = text.strip().splitlines()
    metadata_lines = [line.replace(" ", "") for line in lines if line.startswith("#")]
    data_lines = [line.strip() for line in lines if not line.startswith("#")]
    return metadata_lines, pd.read_csv(
        StringIO("\n".join(data_lines)), sep=r"\s+", header=None
    )


@pytest.mark.asyncio
async def test_xdi_round_trip(tmp_path: Path):
    """
    Steps:

    1. Parse XDI string into DataFrame + dict.
    2. Export as XDI to client.
    3. Read exported XDI into DataFrame + dict.

    Compare result of (3) to result of (1).
    """
    # Write example data file.
    Path(tmp_path / "files").mkdir()
    with open(tmp_path / "files" / "example.xdi", "w") as file:
        file.write(data)
    config = {
        "trees": [
            {
                "tree": "catalog",
                "path": "/",
                "args": {
                    "uri": tmp_path / "catalog.db",
                    "readable_storage": [tmp_path / "files"],
                    "init_if_not_exists": True,
                    "adapters_by_mimetype": {
                        "application/x-xdi": "tiled.examples.xdi:XDIAdapter"
                    },
                },
            }
        ],
        "media_types": {"xdi": {"application/x-xdi": "tiled.examples.xdi:write_xdi"}},
        "file_extensions": {"xdi": "application/x-xdi"},
    }
    with Context.from_app(build_app_from_config(config)) as context:
        client = from_context(context)
        await register(
            client,
            tmp_path / "files",
            adapters_by_mimetype={"application/x-xdi": "tiled.examples.xdi:XDIAdapter"},
            mimetypes_by_file_ext={".xdi": "application/x-xdi"},
        )

        client["example"].export(str(tmp_path / "exported.xdi"))
        actual = Path(tmp_path / "exported.xdi").read_text()
        metadata_actual, df_actual = load_data(actual)
        metadata_expected, df_expected = load_data(data)
        assert df_actual.equals(df_expected)
        assert len(metadata_actual) == len(metadata_expected) and set(
            metadata_actual
        ) == set(metadata_expected)



import copy
import importlib

from .serialization.table import XLSX_MIME_TYPE
from .utils import APACHE_ARROW_FILE_MIME_TYPE, OneShotCachedMap

# This maps MIME types (i.e. file formats) for appropriate Readers.
# OneShotCachedMap is used to defer imports. We don't want to pay up front
# for importing Readers that we will not actually use.
PARQUET_MIMETYPE = "application/x-parquet"
SPARSE_BLOCKS_PARQUET_MIMETYPE = "application/x-parquet;structure=sparse"
ZARR_MIMETYPE = "application/x-zarr"
AWKWARD_BUFFERS_MIMETYPE = "application/x-awkward-buffers"
TILED_SQL_TABLE_MIMETYPE = "application/x-tiled-sql-table"
# TODO: make type[Adapter] after #1047
DEFAULT_ADAPTERS_BY_MIMETYPE = OneShotCachedMap[str, type](
    {
        "image/tiff": lambda: importlib.import_module(
            "..adapters.tiff", __name__
        ).TiffAdapter,
        "multipart/related;type=image/tiff": lambda: importlib.import_module(
            "..adapters.tiff", __name__
        ).TiffSequenceAdapter,
        "image/jpeg": lambda: importlib.import_module(
            "..adapters.jpeg", __name__
        ).JPEGAdapter,
        "multipart/related;type=image/jpeg": lambda: importlib.import_module(
            "..adapters.jpeg", __name__
        ).JPEGSequenceAdapter,
        "text/csv": lambda: importlib.import_module(
            "..adapters.csv", __name__
        ).CSVAdapter,
        "multipart/related;type=text/csv": lambda: importlib.import_module(
            "..adapters.csv", __name__
        ).CSVAdapter,
        # https://www.rfc-editor.org/rfc/rfc4180#section-3
        "text/csv;header=present": lambda: importlib.import_module(
            "..adapters.csv", __name__
        ).CSVAdapter,
        "text/csv;header=absent": lambda: importlib.import_module(
            "..adapters.csv", __name__
        ).CSVArrayAdapter,
        XLSX_MIME_TYPE: lambda: importlib.import_module(
            "..adapters.excel", __name__
        ).ExcelAdapter,
        "application/x-hdf5": lambda: importlib.import_module(
            "..adapters.hdf5", __name__
        ).HDF5Adapter,
        "application/x-netcdf": lambda: importlib.import_module(
            "..adapters.netcdf", __name__
        ).NetCDFAdapter,
        "application/x-npy": lambda: importlib.import_module(
            "..adapters.npy", __name__
        ).NPYAdapter,
        "multipart/related;type=application/x-npy": lambda: importlib.import_module(
            "..adapters.npy", __name__
        ).NPYSequenceAdapter,
        PARQUET_MIMETYPE: lambda: importlib.import_module(
            "..adapters.parquet", __name__
        ).ParquetDatasetAdapter,
        SPARSE_BLOCKS_PARQUET_MIMETYPE: lambda: importlib.import_module(
            "..adapters.sparse_blocks_parquet", __name__
        ).SparseBlocksParquetAdapter,
        ZARR_MIMETYPE: lambda: importlib.import_module(
            "..adapters.zarr", __name__
        ).ZarrAdapter,
        AWKWARD_BUFFERS_MIMETYPE: lambda: importlib.import_module(
            "..adapters.awkward_buffers", __name__
        ).AwkwardBuffersAdapter,
        APACHE_ARROW_FILE_MIME_TYPE: lambda: importlib.import_module(
            "..adapters.arrow", __name__
        ).ArrowAdapter,
        TILED_SQL_TABLE_MIMETYPE: lambda: importlib.import_module(
            "..adapters.sql", __name__
        ).SQLAdapter,
    }
)

DEFAULT_REGISTRATION_ADAPTERS_BY_MIMETYPE = copy.deepcopy(DEFAULT_ADAPTERS_BY_MIMETYPE)

DEFAULT_REGISTRATION_ADAPTERS_BY_MIMETYPE.set(
    "text/csv",
    lambda: importlib.import_module("..adapters.csv", __name__).CSVAdapter,
)


# We can mostly rely on mimetypes.types_map for the common ones
# ('.csv' -> 'text/csv', etc.) but we supplement here for some
# of the more exotic ones that not all platforms know about.
DEFAULT_MIMETYPES_BY_FILE_EXT = {
    # This is the "official" file extension.
    ".h5": "application/x-hdf5",
    # This is the Numpy on-disk format.
    ".npy": "application/x-npy",
    # This is NeXus. We may want to invent a special media type
    # like 'application/x-nexus' for this, but I'll punt that for now.
    # Needs thought about how to encode the various types of NeXus
    # (media type arguments, for example).
    ".nxs": "application/x-hdf5",
    # These are unofficial but common file extensions.
    ".hdf": "application/x-hdf5",
    ".hdf5": "application/x-hdf5",
    # on opensuse csv -> text/x-comma-separated-values
    ".csv": "text/csv",
    ".zarr": ZARR_MIMETYPE,
}



"""
Example code for importing and exporting xdi using tiled

Source: https://github.com/XraySpectroscopy/XAS-Data-Interchange
"""

import collections
import io
import pathlib
import re
from typing import Any, Optional

import dask.dataframe
import pandas as pd

from tiled.adapters.dataframe import DataFrameAdapter
from tiled.adapters.table import TableAdapter
from tiled.adapters.utils import init_adapter_from_catalog
from tiled.catalog.orm import Node
from tiled.structures.core import Spec, StructureFamily
from tiled.structures.data_source import DataSource
from tiled.structures.table import TableStructure
from tiled.type_aliases import JSON
from tiled.utils import path_from_uri


class XDIAdapter(TableAdapter):
    structure_family = StructureFamily.table

    def __init__(
        self,
        data_uri: str,
        structure: Optional[TableStructure] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[list[Spec]] = None,
        **kwargs: Optional[Any],
    ) -> None:
        """Adapter for XDI data"""

        filepath = path_from_uri(data_uri)
        with open(filepath, "r") as file:
            metadata = {}
            fields = collections.defaultdict(dict)

            # if isinstance(f, pathlib.PosixPath):
            #    line = f.read_text().split('\n')[0]
            # else:
            line = file.readline()
            m = re.match(r"#\s*XDI/(\S*)\s*(\S*)?", line)
            if not m:
                raise ValueError(
                    f"not an XDI file, no XDI versioning information in first line\n{line}"
                )

            metadata["xdi_version"] = m[1]
            metadata["extra_version"] = m[2]

            field_end_re = re.compile(r"#\s*/{3,}")
            header_end_re = re.compile(r"#\s*-{3,}")

            has_comments = False

            # read header
            for line in file:
                if line[0] != "#":
                    raise ValueError(f"reached invalid line in header\n{line}")
                if re.match(field_end_re, line):
                    has_comments = True
                    break
                elif re.match(header_end_re, line):
                    break

                try:
                    key, val = line[1:].strip().split(":", 1)
                    val = val.strip()
                    namespace, tag = key.split(".")
                    # TODO coerce to lower case?
                except ValueError:
                    print(f"error processing line\n{line}")
                    raise

                fields[namespace][tag] = val

            if has_comments:
                comments = ""
                for line in file:
                    if re.match(header_end_re, line):
                        break
                    comments += line

                metadata["comments"] = comments

            metadata["fields"] = fields

            line = file.readline()
            if line[0] != "#":
                raise ValueError(f"expected column labels. got\n{line}")
            col_labels = line[1:].split()

            # TODO validate

            df = pd.read_table(file, sep=r"\s+", names=col_labels)

        ddf = dask.dataframe.from_pandas(df, npartitions=1)
        structure = TableStructure.from_dask_dataframe(ddf)

        super().__init__(
            partitions=ddf.partitions,
            structure=structure,
            metadata=metadata,
            specs=(specs or []) + [Spec("xdi", version="1.0")],
        )

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource,
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "XDIAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)

    @classmethod
    def from_uris(
        cls,
        data_uri: str,
        **kwargs: Optional[Any],
    ) -> "XDIAdapter":
        return cls(data_uri, **kwargs)


def read_xdi(data_uri, structure=None, metadata=None, specs=None, access_policy=None):
    "Read XDI-formatted file."
    filepath = path_from_uri(data_uri)
    with open(filepath, "r") as file:
        metadata = {}
        fields = collections.defaultdict(dict)

        # if isinstance(f, pathlib.PosixPath):
        #    line = f.read_text().split('\n')[0]
        # else:
        line = file.readline()
        m = re.match(r"#\s*XDI/(\S*)\s*(\S*)?", line)
        if not m:
            raise ValueError(
                f"not an XDI file, no XDI versioning information in first line\n{line}"
            )

        metadata["xdi_version"] = m[1]
        metadata["extra_version"] = m[2]

        field_end_re = re.compile(r"#\s*/{3,}")
        header_end_re = re.compile(r"#\s*-{3,}")

        has_comments = False

        # read header
        for line in file:
            if line[0] != "#":
                raise ValueError(f"reached invalid line in header\n{line}")
            if re.match(field_end_re, line):
                has_comments = True
                break
            elif re.match(header_end_re, line):
                break

            try:
                key, val = line[1:].strip().split(":", 1)
                val = val.strip()
                namespace, tag = key.split(".")
                # TODO coerce to lower case?
            except ValueError:
                print(f"error processing line\n{line}")
                raise

            fields[namespace][tag] = val

        if has_comments:
            comments = ""
            for line in file:
                if re.match(header_end_re, line):
                    break
                comments += line

            metadata["comments"] = comments

        metadata["fields"] = fields

        line = file.readline()
        if line[0] != "#":
            raise ValueError(f"expected column labels. got\n{line}")
        col_labels = line[1:].split()

        # TODO validate

        df = pd.read_table(file, sep=r"\s+", names=col_labels)

    return DataFrameAdapter.from_pandas(
        df,
        metadata=metadata,
        specs=(specs or []) + [Spec("xdi", version="1.0")],
    )


def write_xdi(mimetype, df, metadata):
    output = io.StringIO()

    xdi_version = metadata.get("xdi_version")
    extra_version = metadata.get("extra_version")

    output.write(f"# XDI/{xdi_version} {extra_version}\n")

    fields = metadata["fields"]
    for namespace, namespace_dict in fields.items():
        for tag, value in namespace_dict.items():
            output.write(f"# {namespace}.{tag}: {value}\n")

    # write column labels
    columns = list(df.columns)
    output.write("# ")
    output.write(" ".join(columns))
    output.write("\n")

    # write data
    df.to_csv(output, header=False, index=False, sep=" ")
    return output.getvalue()


data = """# XDI/1.0 GSE/1.0
# Column.1: energy eV
# Column.2: i0
# Column.3: itrans
# Column.4: mutrans
# Element.edge: K
# Element.symbol: Cu
# Scan.edge_energy: 8980.0
# Mono.name: Si 111
# Mono.d_spacing: 3.13553
# Beamline.name: 13ID
# Beamline.collimation: none
# Beamline.focusing: yes
# Beamline.harmonic_rejection: rhodium-coated mirror
# Facility.name: APS
# Facility.energy: 7.00 GeV
# Facility.xray_source: APS Undulator A
# Scan.start_time: 2001-06-26T22:27:31
# Detector.I0: 10cm  N2
# Detector.I1: 10cm  N2
# Sample.name: Cu
# Sample.prep: Cu metal foil
# GSE.EXTRA:  config 1
# ///
# Cu foil Room Temperature
# measured at beamline 13-ID
#----
# energy i0 itrans mutrans
  8779.0  149013.7  550643.089065  -1.3070486
  8789.0  144864.7  531876.119084  -1.3006104
  8799.0  132978.7  489591.10592  -1.3033816
  8809.0  125444.7  463051.104096  -1.3059724
  8819.0  121324.7  449969.103983  -1.3107085
  8829.0  119447.7  444386.117562  -1.3138152
  8839.0  119100.7  440176.091039  -1.3072055
  8849.0  117707.7  440448.106567  -1.3195882
  8859.0  117754.7  442302.10637  -1.3233895
  8869.0  117428.7  441944.116528  -1.3253521
  8879.0  117383.7  442810.120466  -1.327693
  8889.0  117185.7  443658.11566  -1.3312944"""


def main():
    pathlib.Path("data").mkdir()
    with open("data/example.xdi", "w") as f:
        f.write(data)


if __name__ == "__main__":
    main()



LOGGING_CONFIG = {
    "disable_existing_loggers": False,
    "filters": {
        "principal": {
            "()": "tiled.server.principal_log_filter.PrincipalFilter",
        },
        "correlation_id": {
            "()": "asgi_correlation_id.CorrelationIdFilter",
            "default_value": "-",
            "uuid_length": 16,
        },
    },
    "formatters": {
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "datefmt": "%Y-%m-%dT%H:%M:%S",
            "format": (
                "[%(correlation_id)s] "
                '%(client_addr)s (%(principal)s) - "%(request_line)s" '
                "%(status_code)s"
            ),
            "use_colors": True,
        },
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "datefmt": "%Y-%m-%dT%H:%M:%S",
            "format": "[%(correlation_id)s] %(levelprefix)s %(message)s",
            "use_colors": True,
        },
    },
    "handlers": {
        "access": {
            "class": "logging.StreamHandler",
            "filters": ["principal", "correlation_id"],
            "formatter": "access",
            "stream": "ext://sys.stdout",
        },
        "default": {
            "class": "logging.StreamHandler",
            "filters": ["correlation_id"],
            "formatter": "default",
            "stream": "ext://sys.stderr",
        },
    },
    "loggers": {
        "uvicorn.access": {"handlers": ["access"], "level": "INFO", "propagate": False},
        "uvicorn.error": {
            "handlers": ["default"],
            "level": "INFO",
            "propagate": False,
        },
    },
    "version": 1,
}



"""
Generate the 'links' section of the response JSON.

The links vary by structure family.
"""
from ..structures.core import StructureFamily


def links_for_node(structure_family, structure, base_url, path_str):
    links = {}
    links = LINKS_BY_STRUCTURE_FAMILY[structure_family](
        structure_family, structure, base_url, path_str
    )
    links["self"] = f"{base_url}/metadata/{path_str}"
    return links


def links_for_array(structure_family, structure, base_url, path_str):
    links = {}
    block_template = ",".join(f"{{{index}}}" for index in range(len(structure.shape)))
    links["block"] = f"{base_url}/array/block/{path_str}?block={block_template}"
    links["full"] = f"{base_url}/array/full/{path_str}"
    return links


def links_for_awkward(structure_family, structure, base_url, path_str):
    links = {}
    links["buffers"] = f"{base_url}/awkward/buffers/{path_str}"
    links["full"] = f"{base_url}/awkward/full/{path_str}"
    return links


def links_for_container(structure_family, structure, base_url, path_str):
    links = {}
    links["full"] = f"{base_url}/container/full/{path_str}"
    links["search"] = f"{base_url}/search/{path_str}"
    return links


def links_for_table(structure_family, structure, base_url, path_str):
    links = {}
    links["partition"] = f"{base_url}/table/partition/{path_str}?partition={{index}}"
    links["full"] = f"{base_url}/table/full/{path_str}"
    return links


LINKS_BY_STRUCTURE_FAMILY = {
    StructureFamily.array: links_for_array,
    StructureFamily.awkward: links_for_awkward,
    StructureFamily.container: links_for_container,
    StructureFamily.sparse: links_for_array,  # spare and array are the same
    StructureFamily.table: links_for_table,
}



from typing import Any, Dict, List, Optional

import awkward
import awkward.forms
from numpy.typing import NDArray

from ..storage import FileStorage
from ..structures.awkward import AwkwardStructure
from ..structures.core import Spec, StructureFamily
from ..type_aliases import JSON
from .awkward_directory_container import DirectoryContainer


class AwkwardAdapter:
    structure_family = StructureFamily.awkward
    supported_storage = {FileStorage}

    def __init__(
        self,
        container: DirectoryContainer,
        structure: AwkwardStructure,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        container :
        structure :
        metadata :
        specs :
        """
        self.container = container
        self._metadata = metadata or {}
        self._structure = structure
        self.specs = list(specs or [])

    @classmethod
    def from_array(
        cls,
        array: NDArray[Any],
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> "AwkwardAdapter":
        """

        Parameters
        ----------
        array :
        metadata :
        specs :

        Returns
        -------

        """
        form, length, container = awkward.to_buffers(array)
        structure = AwkwardStructure(length=length, form=form.to_dict())
        return cls(
            container,
            structure,
            metadata=metadata,
            specs=specs,
        )

    def metadata(self) -> JSON:
        """

        Returns
        -------

        """
        return self._metadata

    def read_buffers(self, form_keys: Optional[List[str]] = None) -> Dict[str, bytes]:
        """

        Parameters
        ----------
        form_keys :

        Returns
        -------

        """
        form = awkward.forms.from_dict(self._structure.form)
        keys = [
            key
            for key in form.expected_from_buffers()
            if (form_keys is None)
            or any(key.startswith(form_key) for form_key in form_keys)
        ]
        buffers = {}
        for key in keys:
            buffers[key] = self.container[key]
        return buffers

    def read(self) -> Dict[str, bytes]:
        """

        Returns
        -------

        """
        return dict(self.container)

    def write(self, container: DirectoryContainer) -> None:
        """

        Parameters
        ----------
        container :

        Returns
        -------

        """
        for form_key, value in container.items():
            self.container[form_key] = value

    def structure(self) -> AwkwardStructure:
        return self._structure



from pathlib import Path
from typing import Any, List, Optional, Set, Union

import xarray

from ..catalog.orm import Node
from ..storage import Storage
from ..structures.data_source import DataSource
from ..utils import path_from_uri
from .xarray import DatasetAdapter


def read_netcdf(filepath: Union[str, List[str], Path]) -> DatasetAdapter:
    """

    Parameters
    ----------
    filepath :

    Returns
    -------

    """
    ds = xarray.open_dataset(filepath, decode_times=False)
    return DatasetAdapter.from_dataset(ds)


class NetCDFAdapter:
    supported_storage: Set[type[Storage]] = set()

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[None],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "NetCDFAdapter":
        filepath = path_from_uri(data_source.assets[0].data_uri)

        return read_netcdf(filepath)  # type: ignore



"""Add access_blob column to revisions

Revision ID: a963a6c32a0c
Revises: e05e918092c3
Create Date: 2025-08-12 16:22:07.744963

"""
import sqlalchemy as sa
from alembic import op

from tiled.catalog.orm import JSONVariant

# revision identifiers, used by Alembic.
revision = "a963a6c32a0c"
down_revision = "e05e918092c3"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "revisions",
        sa.Column("access_blob", JSONVariant, nullable=False, server_default="{}"),
    )


def downgrade():
    op.drop_column("revisions", "access_blob")



from collections import OrderedDict

from ..utils import Sentinel

DELETE_KEY = Sentinel("DELETE_KEY")


def apply_update_patch(*objs, **kw):
    # adapted from https://github.com/OpenDataServices/json-merge-patch
    result = objs[0]
    for obj in objs[1:]:
        result = _update_obj(result, obj, kw.get("position"))
    return result


def _update_obj(result, obj, position=None):
    # adapted from https://github.com/OpenDataServices/json-merge-patch
    if not isinstance(result, dict):
        result = OrderedDict() if position else {}

    if not isinstance(obj, dict):
        return obj

    if position:
        if position not in ("first", "last"):
            raise ValueError("position can either be first or last")
        if not isinstance(result, OrderedDict) or not isinstance(obj, OrderedDict):
            raise ValueError("If using position all dicts need to be OrderedDicts")

    for key, value in obj.items():
        if isinstance(value, dict):
            target = result.get(key)
            if isinstance(target, dict):
                _update_obj(target, value, position)
                continue
            result[key] = OrderedDict() if position else {}
            if position and position == "first":
                result.move_to_end(key, False)
            _update_obj(result[key], value, position)
            continue
        if value is DELETE_KEY:
            result.pop(key, None)
            continue
        if key not in result and position == "first":
            result[key] = value
            result.move_to_end(key, False)
        else:
            result[key] = value

    return result



"""Redefine the catalog tree using a closure table

Revision ID: e05e918092c3
Revises: 7809873ea2c7
Create Date: 2025-06-16 10:38:46.797381

This revision introduces a closure table to represent the hierarchical structure of nodes in the catalog.

Tree structure:
---------------

root
├── A
│   ├── C
│   │   ├── E
│   │   ├── F
│   │   └── G
│   └── D
└── B

Before:
-------

> SELECT * FROM nodes;

id  key  ancestors  structure_family  metadata   ...
--  ---  ---------  ----------------  --------
1   A    []         container         {}
2   B    []         container         {}
3   C    ["A"]      container         {}
4   D    ["A"]      container         {}
5   E    ["A","C"]  container         {}
6   F    ["A","C"]  container         {}
7   G    ["A","C"]  container         {}

After:
------

> SELECT * FROM nodes;

id  parent  key  structure_family  metadata   ...
--  ------  ---  ----------------  --------
0                container         {}         <-- root node (always present)
1   0       A    container         {}
2   0       B    container         {}
3   1       C    container         {}
4   1       D    container         {}
5   3       E    container         {}
6   3       F    container         {}
7   3       G    container         {}

> SELECT * FROM nodes_closure;

ancestor  descendant  depth
--------  ----------  -----
0         0           0
1         1           0
0         1           1
2         2           0
0         2           1
3         3           0
1         3           1
0         3           2
4         4           0
1         4           1
0         4           2
5         5           0
3         5           1
1         5           2
0         5           3
6         6           0
3         6           1
1         6           2
0         6           3
7         7           0
3         7           1
1         7           2
0         7           3

A new node with `id = 0` and no parent representing the root of the tree is added to the nodes table.
The `parent` column is introduced to represent the parent-child relationships between nodes.
The `ancestors` column is removed, as it is no longer needed with the closure table structure.

"""
import logging

import sqlalchemy as sa
from alembic import op

from tiled.catalog.orm import JSONVariant

# revision identifiers, used by Alembic.
revision = "e05e918092c3"
down_revision = "7809873ea2c7"
branch_labels = None
depends_on = None

logger = logging.getLogger(__name__)
logger.setLevel("INFO")
handler = logging.StreamHandler()
handler.setLevel("DEBUG")
handler.setFormatter(logging.Formatter("%(message)s"))
logger.addHandler(handler)


def upgrade():
    connection = op.get_bind()

    logger.info("Starting migration to add closure table...")

    # 1. Add the 'parent' column and the foreign key to the 'nodes' table. Use batch mode, so it works for SQLite.
    with op.batch_alter_table("nodes", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("parent", sa.Integer(), nullable=True, index=True)
        )
        batch_op.create_foreign_key("fk_nodes_parent", "nodes", ["parent"], ["id"])
        batch_op.drop_constraint("key_ancestors_unique_constraint", type_="unique")
    logger.info("Added 'parent' column and foreign key to 'nodes' table.")

    # 2. Create the 'nodes_closure' table and create the uniqueness constraint
    op.create_table(
        "nodes_closure",
        sa.Column(
            "ancestor", sa.Integer(), sa.ForeignKey("nodes.id"), primary_key=True
        ),
        sa.Column(
            "descendant", sa.Integer(), sa.ForeignKey("nodes.id"), primary_key=True
        ),
        sa.Column("depth", sa.Integer(), nullable=False),
    )
    with op.batch_alter_table("nodes_closure", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "ancestor_descendant_unique_constraint", ["ancestor", "descendant"]
        )
    op.create_index("idx_nodes_closure_ancestor", "nodes_closure", ["ancestor"])
    op.create_index("idx_nodes_closure_descendant", "nodes_closure", ["descendant"])
    logger.info("Created 'nodes_closure' table with uniqueness constraint.")

    # 3. Insert the explicit root node (id=0, key='') with no parent
    connection.execute(
        sa.text(
            """
        INSERT INTO nodes(id, key, parent, ancestors, structure_family, metadata, specs, access_blob)
        VALUES (0, '', NULL, NULL, 'container', '{}', '[]', '{}');
        """
        )
    )
    logger.info("Inserted root node with id=0.")

    # 4. Insert self-referential records into nodes_closure for each node, including the "root" node
    connection.execute(
        sa.text(
            """
        INSERT INTO nodes_closure(ancestor, descendant, depth)
        SELECT id, id, 0 FROM nodes;
        """
        )
    )
    logger.info("Inserted self-referential records into 'nodes_closure' for each node.")

    # 5. Populate the 'parent' column of the 'nodes' table based on the 'ancestors' column
    json_len_func = (
        "jsonb_array_length"
        if connection.engine.dialect.name == "postgresql"
        else "json_array_length"
    )
    max_depth = (
        connection.execute(
            sa.text(
                f"SELECT MAX({json_len_func}(ancestors)) FROM nodes WHERE ancestors IS NOT NULL;"
            )
        ).scalar()
        or 0
    )
    logger.info(f"Maximum depth of ancestors found: {max_depth}")

    # 6. Initialize the parent of each node as 0 (the 'root' node) and set 'depth' in the closure table
    connection.execute(
        sa.text("UPDATE nodes SET parent = 0 WHERE ancestors IS NOT NULL;")
    )
    connection.execute(
        sa.text(
            f"""
        INSERT INTO nodes_closure(ancestor, descendant, depth)
        SELECT 0, id, {json_len_func}(ancestors) + 1 FROM nodes WHERE ancestors IS NOT NULL;
        """
        )
    )
    logger.info(
        "Initialized parent of each node as 0 and set depth in 'nodes_closure'."
    )

    # 7. Update the 'parent' column recursively
    for depth in range(max_depth):
        condition_statement = (
            f"parent.key = child.ancestors::json->>{depth}"
            if connection.engine.dialect.name == "postgresql"
            else f"parent.key = json_extract(child.ancestors, '$[{depth}]')"
        )
        connection.execute(
            sa.text(
                f"""
            UPDATE nodes AS child
            SET parent = parent.id
            FROM nodes AS parent
            WHERE {json_len_func}(child.ancestors) >= {depth + 1}
            AND {condition_statement}
            AND parent.parent = child.parent
        """
            )
        )

        # Populate the 'nodes_closure' table   (possibly use ON CONFLICT DO NOTHING ?)
        connection.execute(
            sa.text(
                f"""
            INSERT INTO nodes_closure (ancestor, descendant, depth)
            SELECT parent.id, child.id, {json_len_func}(child.ancestors) - {depth}
            FROM nodes AS child
            JOIN nodes AS parent ON parent.id = child.parent
            WHERE {json_len_func}(child.ancestors) >= {depth + 1};
        """
            )
        )
        logger.info(
            f"Updated 'parent' column and 'nodes_closure' for depth {depth + 1}."
        )
    logger.info("Completed updating 'parent' column recursively.")

    # 8. Update index in the 'nodes' table: drop old, add new
    op.drop_index("top_level_metadata", table_name="nodes")
    op.create_index(
        "top_level_metadata",
        "nodes",
        ["parent", "time_created", "id", "metadata", "access_blob"],
        postgresql_using="gin",
    )
    logger.info("Updated index in the 'nodes' table.")

    # 9. Create constraint to ensure uniqueness of (key, parent) pairs
    with op.batch_alter_table("nodes", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "key_parent_unique_constraint", ["key", "parent"]
        )
    logger.info("Created unique constraint on (key, parent) pairs in 'nodes' table.")

    # 10. Drop the 'ancestors' column from the 'nodes' table
    op.drop_column("nodes", "ancestors")
    logger.info("Dropped 'ancestors' column from the 'nodes' table.")

    # 11. Add triggers to maintain the closure table
    if connection.engine.dialect.name == "sqlite":
        # Create a trigger to update the closure table when INSERTING a new node
        connection.execute(
            sa.text(
                """
CREATE TRIGGER update_closure_table_when_inserting
AFTER INSERT ON nodes
BEGIN
    INSERT INTO nodes_closure(ancestor, descendant, depth)
    SELECT NEW.id, NEW.id, 0;
    INSERT INTO nodes_closure(ancestor, descendant, depth)
    SELECT p.ancestor, c.descendant, p.depth+c.depth+1
    FROM nodes_closure p, nodes_closure c
    WHERE p.descendant=NEW.parent and c.ancestor=NEW.id;
END"""
            )
        )

        # Create a trigger to update the closure table when DELETING a node
        connection.execute(
            sa.text(
                """
CREATE TRIGGER update_closure_table_when_deleting
BEFORE DELETE ON nodes
BEGIN
    DELETE FROM nodes_closure
    WHERE (ancestor, descendant) IN (
    SELECT p.ancestor, c.descendant
    FROM nodes_closure p, nodes_closure c
    WHERE (p.descendant=OLD.parent OR p.descendant=OLD.id) AND c.ancestor=OLD.id);
END"""
            )
        )

    elif connection.engine.dialect.name == "postgresql":
        # Create function and trigger to update the closure table when INSERTING a new node
        connection.execute(
            sa.text(
                """
CREATE OR REPLACE FUNCTION update_closure_table_when_inserting()
RETURNS TRIGGER AS $$
BEGIN
    INSERT INTO nodes_closure(ancestor, descendant, depth)
    VALUES (NEW.id, NEW.id, 0);
    INSERT INTO nodes_closure(ancestor, descendant, depth)
    SELECT p.ancestor, c.descendant, p.depth + c.depth + 1
    FROM nodes_closure p, nodes_closure c
    WHERE p.descendant = NEW.parent AND c.ancestor = NEW.id;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;
"""
            )
        )

        connection.execute(
            sa.text(
                """
CREATE TRIGGER update_closure_table_when_inserting
AFTER INSERT ON nodes
FOR EACH ROW
EXECUTE FUNCTION update_closure_table_when_inserting();
"""
            )
        )

        # Create function and trigger to update the closure table when DELETING a node
        connection.execute(
            sa.text(
                """
CREATE OR REPLACE FUNCTION update_closure_table_when_deleting()
RETURNS TRIGGER AS $$
BEGIN
    DELETE FROM nodes_closure
    WHERE (ancestor, descendant) IN (
        SELECT p.ancestor, c.descendant
        FROM nodes_closure p, nodes_closure c
        WHERE (p.descendant = OLD.parent OR p.descendant = OLD.id)
        AND c.ancestor = OLD.id
    );
    RETURN OLD;
END;
$$ LANGUAGE plpgsql;
"""
            )
        )

        connection.execute(
            sa.text(
                """
CREATE TRIGGER update_closure_table_when_deleting
BEFORE DELETE ON nodes
FOR EACH ROW
EXECUTE FUNCTION update_closure_table_when_deleting();
"""
            )
        )
    logger.info("Added triggers to maintain the closure table.")
    logger.info("Migration to add closure table completed successfully.")


def downgrade():
    connection = op.get_bind()

    logger.info("Starting downgrade to remove closure table...")

    # 1. Drop triggers and functions for maintaining the closure table
    if connection.engine.dialect.name == "sqlite":
        connection.execute(
            sa.text("DROP TRIGGER IF EXISTS update_closure_table_when_inserting")
        )
        connection.execute(
            sa.text("DROP TRIGGER IF EXISTS update_closure_table_when_deleting")
        )
    elif connection.engine.dialect.name == "postgresql":
        connection.execute(
            sa.text(
                "DROP TRIGGER IF EXISTS update_closure_table_when_inserting ON nodes"
            )
        )
        connection.execute(
            sa.text(
                "DROP TRIGGER IF EXISTS update_closure_table_when_deleting ON nodes"
            )
        )
        connection.execute(
            sa.text("DROP FUNCTION IF EXISTS update_closure_table_when_inserting")
        )
        connection.execute(
            sa.text("DROP FUNCTION IF EXISTS update_closure_table_when_deleting")
        )
    logger.info("Dropped triggers and functions for maintaining the closure table.")

    # 2. Re-add the 'ancestors' column to the 'nodes' table
    op.add_column("nodes", sa.Column("ancestors", JSONVariant, nullable=True))

    # 3. Reconstruct 'ancestors' for each node from 'parent'. Skip the root node (id=0)
    # In PostgreSQL, jsonb_agg does not support ORDER BY directly in SELECT.
    # SQLite’s json_group_array only allows ORDER BY outside the function.
    select_stmt = (
        "SELECT jsonb_agg(n.key ORDER BY nc.depth DESC)"
        if connection.engine.dialect.name == "postgresql"
        else "SELECT json_group_array(n.key)"
    )
    connection.execute(
        sa.text(
            f"""
            UPDATE nodes
            SET ancestors = COALESCE((
                {select_stmt}
                FROM nodes_closure nc
                JOIN nodes n ON nc.ancestor = n.id
                WHERE nc.descendant = nodes.id AND nc.depth > 0 AND nc.ancestor != 0
                {"" if connection.engine.dialect.name == "postgresql" else "ORDER BY nc.depth DESC"}
            ), '[]'{"::jsonb" if connection.engine.dialect.name == "postgresql" else ""})
            WHERE id != 0;
            """
        )
    )
    logger.info("Reconstructed 'ancestors' for each node from 'parent'.")

    # 4. Drop the closure table
    op.drop_table("nodes_closure")
    logger.info("Dropped 'nodes_closure' table.")

    # 5. Restore the old index, drop the new one
    op.drop_index("top_level_metadata", table_name="nodes")
    op.create_index(
        "top_level_metadata",
        "nodes",
        ["time_created", "id", "ancestors", "metadata", "access_blob"],
        postgresql_using="gin",
    )
    logger.info("Restored old index in the 'nodes' table.")

    # 6. Drop the 'parent' column and related foreign key/unique constraint
    with op.batch_alter_table("nodes", schema=None) as batch_op:
        batch_op.drop_index("idx_nodes_parent")
        batch_op.drop_constraint("fk_nodes_parent", type_="foreignkey")
        batch_op.drop_constraint("key_parent_unique_constraint", type_="unique")
        batch_op.create_unique_constraint(
            "key_ancestors_unique_constraint", ["key", "ancestors"]
        )
        batch_op.drop_column("parent")
    logger.info(
        "Dropped 'parent' column and restored 'key_ancestors_unique_constraint'."
    )

    # 7. Remove the explicit root node from the 'nodes' table
    connection.execute(sa.text("DELETE FROM nodes WHERE id = 0"))
    logger.info("Removed the root node from the 'nodes' table.")
    logger.info("Downgrade to remove closure table completed successfully.")



import time
from typing import Iterable, Optional, Union
from urllib.parse import parse_qs, urlparse

from ..structures.core import StructureFamily
from .container import LENGTH_CACHE_TTL, Container
from .utils import MSGPACK_MIME_TYPE, handle_error, retry_context


class CompositeClient(Container):
    def get_contents(self, maxlen=None, include_metadata=False):
        result = {}
        next_page_url = f"{self.item['links']['search']}"
        while (next_page_url is not None) or (
            maxlen is not None and len(result) < maxlen
        ):
            for attempt in retry_context():
                with attempt:
                    content = handle_error(
                        self.context.http_client.get(
                            next_page_url,
                            headers={"Accept": MSGPACK_MIME_TYPE},
                            params={
                                **parse_qs(urlparse(next_page_url).query),
                                **self._queries_as_params,
                            }
                            | ({} if include_metadata else {"select_metadata": False})
                            | (
                                {}
                                if not self._include_data_sources
                                else {"include_data_sources": True}
                            ),
                        )
                    ).json()
            result.update({item["id"]: item for item in content["data"]})

            next_page_url = content["links"]["next"]

        return result

    @property
    def _flat_keys_mapping(self):
        result = {}
        for key, item in self.get_contents().items():
            if item["attributes"]["structure_family"] == StructureFamily.table:
                for col in item["attributes"]["structure"]["columns"]:
                    result[col] = item["id"] + "/" + col
            else:
                result[item["id"]] = item["id"]

        self._cached_len = (len(result), time.monotonic() + LENGTH_CACHE_TTL)

        return result

    @property
    def base(self):
        "Return the base Container client instead of a CompositeClient"
        return Container(
            self.context, item=self.item, structure_clients=self.structure_clients
        )

    def _keys_slice(self, start, stop, direction, _ignore_inlined_contents=False):
        yield from self._flat_keys_mapping.keys()

    def _items_slice(self, start, stop, direction, _ignore_inlined_contents=False):
        for key in self._flat_keys_mapping.keys():
            yield key, self[key]

    def __iter__(self):
        yield from self._keys_slice(0, None, 1)

    def __len__(self):
        if self._cached_len is not None:
            length, deadline = self._cached_len
            if time.monotonic() < deadline:
                # Used the cached value and do not make any request.
                return length

        return len(self._flat_keys_mapping)

    def __getitem__(self, key: str, _ignore_inlined_contents=False):
        if isinstance(key, tuple):
            key = "/".join(key)
        if key in self._flat_keys_mapping:
            key = self._flat_keys_mapping[key]
        else:
            raise KeyError(
                f"Key '{key}' not found. If it refers to a table, access it via "
                f"the base Container client using `.base['{key}']` instead."
            )

        return super().__getitem__(key, _ignore_inlined_contents)

    def __contains__(self, key):
        return key in self._flat_keys_mapping.keys()

    def create_container(self, key=None, *, metadata=None, specs=None):
        """Composite nodes can not include nested containers by design."""
        raise NotImplementedError("Cannot create a container within a composite node.")

    def delete_contents(
        self,
        keys: Optional[Union[str, Iterable[str]]] = None,
        external_only: bool = True,
    ) -> "CompositeClient":
        """Delete the contents of this Composite node.

        Only arrays or entire tables, not individual columns, can be deleted.

        Parameters
        ----------
        keys : str or list of str
            The key(s) to delete. If a list, all keys in the list will be deleted.
            If None (default), delete all contents.
        external_only : bool, optional
            If True, only delete externally-managed data. Defaults to True.
        """

        parts = set(self.base.keys())
        if keys is None:
            keys = parts
        keys = [keys] if isinstance(keys, str) else keys
        extra_keys = set(keys).difference(parts)
        if extra_keys:
            raise KeyError(
                f"Keys {extra_keys} not found in composite node parts. "
                "If the keys reference column names of a constituent "
                "table, deleting them is not supported. Use the `.base` "
                "accessor to the Container client to delete the entire table."
            )

        return super().delete_contents(keys, external_only=external_only)

    def read(self, variables=None, dim0=None):
        """Download the contents of a composite node as an xarray.Dataset.

        Parameters
        ----------
        variables (list, optional) : List of variable names to read. If None, all
            variables are read. Defaults to None.
        dim0 (str, optional) : Name of the dimension to use for the first dimension;
            if None (default), each array will have its own dimension name. The dims tuple,
            if specified in the structure, takes precedence over this.

        Returns
        -------
        xarray.Dataset: The dataset containing the requested variables.
        """
        import pandas
        import xarray

        array_dims = {}
        data_vars = {}
        for part, item in self.get_contents().items():
            # Read all or selective arrays/columns.
            if item["attributes"]["structure_family"] in {
                StructureFamily.array,
                StructureFamily.sparse,
            }:
                if (variables is None) or (part in variables):
                    array_client = self.base[part]
                    data_vars[part] = array_client.read()  # [Dask]ArrayClient
                    array_dims[part] = array_client.dims
            elif item["attributes"]["structure_family"] == StructureFamily.awkward:
                if (variables is None) or (part in variables):
                    try:
                        data_vars[part] = self.base[part].read().to_numpy()
                    except ValueError as e:
                        raise ValueError(
                            f"Failed to convert awkward array to numpy: {e}"
                        ) from e
            elif item["attributes"]["structure_family"] == StructureFamily.table:
                # For now, greedily load tabular data. We cannot know the shape
                # of the columns without reading them. Future work may enable
                # this to be lazy.
                table_client = self.base[part]
                columns = set(variables or table_client.columns).intersection(
                    table_client.columns
                )
                df = table_client.read(list(columns))
                for column in columns:
                    data_vars[column] = df[column].values
                    # Convert (experimental) pandas.StringDtype to numpy's unicode string dtype
                    if isinstance(data_vars[column].dtype, pandas.StringDtype):
                        data_vars[column] = data_vars[column].astype("U")
            else:
                raise ValueError(
                    f"Unsupported structure family: {item['attributes']['structure_family']}"
                )

        # Create xarray.Dataset from the data_vars dictionary
        is_dim0_consistent = (
            len(
                set(
                    [
                        arr.shape[0]
                        for var_name, arr in data_vars.items()
                        if arr.ndim > 0 and not array_dims.get(var_name)
                    ]
                )
            )
            <= 1
        )
        if dim0 is not None and not is_dim0_consistent:
            raise ValueError(
                "Cannot specify dim0 when the arrays have different left-most dimensions."
            )

        for var_name, arr in data_vars.items():
            if is_dim0_consistent:
                dims = (dim0 or "dim0",) + tuple(
                    f"{var_name}_dim{i+1}" for i in range(len(arr.shape) - 1)
                )
            else:
                dims = tuple(f"{var_name}_dim{i}" for i in range(len(arr.shape)))
            data_vars[var_name] = array_dims.get(var_name) or dims, arr

        return xarray.Dataset(data_vars=data_vars)

    def new(
        self,
        structure_family,
        data_sources,
        *,
        key=None,
        metadata=None,
        specs=None,
        access_tags=None,
    ):
        if key in self.keys():
            raise ValueError(f"Key '{key}' already exists in the composite node.")

        return super().new(
            structure_family,
            data_sources,
            key=key,
            metadata=metadata,
            specs=specs,
            access_tags=access_tags,
        )

    def write_dataframe(
        self, dataframe, *, key=None, metadata=None, specs=None, access_tags=None
    ):
        if set(self.keys()).intersection(dataframe.columns):
            raise ValueError(
                "DataFrame columns must not overlap with existing keys in the composite node."
            )
        return super().write_dataframe(
            dataframe, key=key, metadata=metadata, specs=specs, access_tags=access_tags
        )



"""Restore server_default for time_created

Revision ID: 0c705a02954c
Revises: d88e91ea03f9
Create Date: 2025-02-19 13:03:35.067755

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0c705a02954c"
down_revision = "d88e91ea03f9"
branch_labels = None
depends_on = None

tables = ["principals", "identities", "roles", "api_keys", "sessions"]


def upgrade():
    connection = op.get_bind()
    if connection.engine.dialect.name == "sqlite":
        return
    for table in tables:
        connection.execute(
            sa.text(
                f"""
ALTER TABLE {table}
ALTER COLUMN time_created
SET DEFAULT CURRENT_TIMESTAMP;
            """
            )
        )


def downgrade():
    # No action required
    pass



import itertools
from collections.abc import Mapping
from typing import Any, Iterator, List, Optional

import xarray

from ..structures.core import Spec
from .array import ArrayAdapter
from .mapping import MapAdapter


class DatasetAdapter(MapAdapter):
    """
    Wrap an xarray.Dataset
    """

    @classmethod
    def from_dataset(
        cls,
        dataset: Any,
        *,
        specs: Optional[List[Spec]] = None,
    ) -> "DatasetAdapter":
        """

        Parameters
        ----------
        dataset :
        specs :

        Returns
        -------

        """
        mapping = _DatasetMap(dataset)
        specs = specs or []
        if "xarray_dataset" not in [spec.name for spec in specs]:
            specs.append(Spec("xarray_dataset"))
        return cls(
            mapping,
            metadata={"attrs": dataset.attrs},
            specs=specs,
        )

    def __init__(
        self,
        mapping: Any,
        *args: Any,
        specs: Optional[List[Spec]] = None,
        **kwargs: Any,
    ) -> None:
        """

        Parameters
        ----------
        mapping :
        args :
        specs :
        kwargs :
        """
        if isinstance(mapping, xarray.Dataset):
            raise TypeError(
                "Use DatasetAdapter.from_dataset(...), not DatasetAdapter(...)."
            )
        super().__init__(mapping, *args, specs=specs, **kwargs)

    def inlined_contents_enabled(self, depth: int) -> bool:
        """

        Parameters
        ----------
        depth :

        Returns
        -------

        """
        # Tell the server to in-line the description of each array
        # (i.e. data_vars and coords) to avoid latency of a second
        # request.
        return True


class _DatasetMap(Mapping[str, Any]):
    def __init__(self, dataset: Any) -> None:
        """

        Parameters
        ----------
        dataset :
        """
        self._dataset = dataset

    def __len__(self) -> int:
        """

        Returns
        -------

        """
        return len(self._dataset.data_vars) + len(self._dataset.coords)

    def __iter__(self) -> Iterator[Any]:
        """

        Returns
        -------

        """
        yield from itertools.chain(self._dataset.data_vars, self._dataset.coords)

    def __getitem__(self, key: str) -> ArrayAdapter:
        """

        Parameters
        ----------
        key :

        Returns
        -------

        """
        data_array = self._dataset[key]
        if key in self._dataset.coords:
            spec = Spec("xarray_coord")
        else:
            spec = Spec("xarray_data_var")
        return ArrayAdapter.from_array(
            data_array.data,
            metadata={"attrs": data_array.attrs},
            dims=data_array.dims,
            specs=[spec],
        )



import asyncio
import re
from typing import List

import typer


def register(
    uri: str = typer.Argument(..., help="URL to Tiled node to register on"),
    filepath: str = typer.Argument(..., help="A file or directory to register"),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help=("Log details of directory traversal and file registration."),
    ),
    watch: bool = typer.Option(
        False,
        "--watch",
        "-w",
        help="Update catalog when files are added, removed, or changed.",
    ),
    prefix: str = typer.Option(
        "/", help="Location within the catalog's namespace to register these files"
    ),
    keep_ext: bool = typer.Option(
        False,
        "--keep-ext",
        help=(
            "Serve a file like 'measurements.csv' as its full filepath with extension, "
            "instead of the default which would serve it as 'measurements'. "
            "This is discouraged because it leaks details about the storage "
            "format to the client, such that changing the storage in the future "
            "may break user (client-side) code."
        ),
    ),
    ext: List[str] = typer.Option(
        None,
        "--ext",
        help=(
            "Support custom file extension, mapping it to a known mimetype. "
            "Spell like '.tif=image/tiff'. Include the leading '.' in the file "
            "extension."
        ),
    ),
    mimetype_detection_hook: str = typer.Option(
        None,
        "--mimetype-hook",
        help=(
            "ADVANCED: Custom mimetype detection Python function. "
            "Expected interface: detect_mimetype(filepath, mimetype) -> mimetype "
            "Specify here as 'package.module:function'"
        ),
    ),
    adapters: List[str] = typer.Option(
        None,
        "--adapter",
        help=(
            "ADVANCED: Custom Tiled Adapter for reading a given format"
            "Specify here as 'mimetype=package.module:function'"
        ),
    ),
    walkers: List[str] = typer.Option(
        None,
        "--walker",
        help=(
            "ADVANCED: Custom Tiled Walker for traversing directories and "
            "grouping files. This is used in conjunction with Adapters that operate "
            "on groups of files. "
            "Specify here as 'package.module:function'"
        ),
    ),
    api_key: str = typer.Option(
        None,
        "--api-key",
    ),
):
    if keep_ext:
        from ..client.register import identity

        key_from_filename = identity
    else:
        key_from_filename = None
    mimetypes_by_file_ext = {}
    EXT_PATTERN = re.compile(r"(.*) *= *(.*)")
    for item in ext or []:
        match = EXT_PATTERN.match(item)
        if match is None:
            raise ValueError(
                f"Failed parsing --ext option {item}, expected format '.ext=mimetype'"
            )
        ext, mimetype = match.groups()
        mimetypes_by_file_ext[ext] = mimetype
    adapters_by_mimetype = {}
    ADAPTER_PATTERN = re.compile(r"(.*) *= *(.*)")
    for item in adapters or []:
        match = ADAPTER_PATTERN.match(item)
        if match is None:
            raise ValueError(
                f"Failed parsing --adapter option {item}, expected format 'mimetype=package.module:obj'"
            )
        mimetype, obj_ref = match.groups()
        adapters_by_mimetype[mimetype] = obj_ref

    from ..client import from_uri

    client_node = from_uri(uri, api_key=api_key)

    from logging import StreamHandler

    from ..client.register import logger as register_logger
    from ..client.register import register
    from ..client.register import watch as watch_

    if verbose:
        register_logger.addHandler(StreamHandler())
        register_logger.setLevel("INFO")
    if watch:
        asyncio.run(
            watch_(
                client_node,
                filepath,
                prefix=prefix,
                mimetype_detection_hook=mimetype_detection_hook,
                mimetypes_by_file_ext=mimetypes_by_file_ext,
                adapters_by_mimetype=adapters_by_mimetype,
                walkers=walkers,
                key_from_filename=key_from_filename,
            )
        )
    else:
        asyncio.run(
            register(
                client_node,
                filepath,
                prefix=prefix,
                mimetype_detection_hook=mimetype_detection_hook,
                mimetypes_by_file_ext=mimetypes_by_file_ext,
                adapters_by_mimetype=adapters_by_mimetype,
                walkers=walkers,
                key_from_filename=key_from_filename,
            )
        )



"""Add 'register' scope to default roles

Revision ID: 13024b8a6b74
Revises: 769180ce732e
Create Date: 2024-02-21 07:49:08.993168

"""
from alembic import op
from sqlalchemy.orm.session import Session

from tiled.authn_database.orm import Role

# revision identifiers, used by Alembic.
revision = "13024b8a6b74"
down_revision = "769180ce732e"
branch_labels = None
depends_on = None


ROLES = ["admin"]
NEW_SCOPES = ["register"]


def upgrade():
    """
    Add new scopes to Roles.
    """
    connection = op.get_bind()
    with Session(bind=connection) as db:
        for role_name in ROLES:
            role = db.query(Role).filter(Role.name == role_name).first()
            scopes = role.scopes.copy()
            scopes.extend(NEW_SCOPES)
            role.scopes = scopes
            db.commit()


def downgrade():
    """
    Remove new scopes from Roles, if present.
    """
    connection = op.get_bind()
    with Session(bind=connection) as db:
        for role_name in ROLES:
            role = db.query(Role).filter(Role.name == role_name).first()
            scopes = role.scopes.copy()
            for scope in NEW_SCOPES:
                if scope in scopes:
                    scopes.remove(scope)
            role.scopes = scopes
            db.commit()



import collections
import dataclasses
import logging
import mimetypes
import re
from functools import partial
from pathlib import Path

import anyio
import httpx
import watchfiles

from ..mimetypes import (
    DEFAULT_MIMETYPES_BY_FILE_EXT,
    DEFAULT_REGISTRATION_ADAPTERS_BY_MIMETYPE,
)
from ..structures.core import StructureFamily
from ..structures.data_source import Asset, DataSource, Management
from ..utils import ensure_uri, import_object
from .utils import ClientError

logger = logging.getLogger(__name__)


def strip_suffixes(filename):
    """
    For use with key_from_filename parameter.

    This gives the 'base' as the key.

    >>> strip_suffixes("a.tif")
    "a"

    >>> strip_suffixes("thing.tar.gz")
    "thing"
    """
    path = Path(filename)
    # You would think there would be a method for this, but there is not.
    if len(path.suffixes):
        return str(path)[: -sum([len(s) for s in path.suffixes])]
    else:
        return filename


def identity(filename):
    """
    For use with key_from_filename parameter.

    This give the full filename (with suffixes) as the key.
    """
    return filename


def default_filter(path):
    "By default, ignore only hidden files."
    return not path.name.startswith(".")


def resolve_mimetype(path, mimetypes_by_file_ext, mimetype_detection_hook=None):
    """
    Given a filepath (file or directory) detect the mimetype.

    If no mimetype could be resolved, return None.
    """
    # First, try to infer the mimetype from the file extension.
    # For compound suffixes like '.u1.strict_disabled.avif' (a real example)
    # consider in order:
    # '.u1.strict_disabled.avif'
    # '.strict_disabled.avif'
    # '.avif'
    for i in range(len(path.suffixes)):
        ext = "".join(path.suffixes[i:])  # e.g. ".h5" or ".tar.gz"
        if ext in mimetypes_by_file_ext:
            mimetype = mimetypes_by_file_ext[ext]
            break
    else:
        # Use the Python's built-in facility for guessing mimetype
        # from file extension. This loads data about mimetypes from
        # the operating system the first time it is used.
        mimetype, _ = mimetypes.guess_type(str(path))
    # Finally, user-specified function has the opportunity to
    # look at more than just the file extension. This gets access to the full
    # path, so it can consider the file name and even open the file. It is also
    # passed the mimetype determined above, or None if no match was found.
    if mimetype_detection_hook is not None:
        mimetype = mimetype_detection_hook(path, mimetype)
    return mimetype


@dataclasses.dataclass(frozen=True)
class Settings:
    adapters_by_mimetype: dict
    mimetypes_by_file_ext: dict
    mimetype_detection_hook: callable
    key_from_filename: callable
    filter: callable

    @classmethod
    def init(
        cls,
        adapters_by_mimetype=None,
        mimetypes_by_file_ext=None,
        mimetype_detection_hook=None,
        key_from_filename=None,
        filter=None,
    ):
        # If parameters come from a configuration file, they
        # are given as importable strings, like "package.module:Reader".
        adapters_by_mimetype = adapters_by_mimetype or {}
        for key, value in list((adapters_by_mimetype).items()):
            if isinstance(value, str):
                adapters_by_mimetype[key] = import_object(value)
        merged_adapters_by_mimetype = collections.ChainMap(
            adapters_by_mimetype, DEFAULT_REGISTRATION_ADAPTERS_BY_MIMETYPE
        )
        if isinstance(key_from_filename, str):
            key_from_filename = import_object(key_from_filename)
        elif key_from_filename is None:
            key_from_filename = strip_suffixes
        if mimetype_detection_hook is not None:
            mimetype_detection_hook = import_object(mimetype_detection_hook)
        merged_mimetypes_by_file_ext = collections.ChainMap(
            mimetypes_by_file_ext or {}, DEFAULT_MIMETYPES_BY_FILE_EXT
        )
        if filter is None:
            filter = default_filter
        if isinstance(filter, str):
            filter = import_object(filter)
        return cls(
            adapters_by_mimetype=merged_adapters_by_mimetype,
            mimetypes_by_file_ext=merged_mimetypes_by_file_ext,
            mimetype_detection_hook=mimetype_detection_hook,
            key_from_filename=key_from_filename,
            filter=filter,
        )


async def register(
    node,
    path,
    prefix="/",
    walkers=None,
    adapters_by_mimetype=None,
    mimetypes_by_file_ext=None,
    mimetype_detection_hook=None,
    key_from_filename=None,
    filter=None,
    overwrite=True,
):
    "Register a file or directory (recursively)."
    settings = Settings.init(
        adapters_by_mimetype=adapters_by_mimetype,
        mimetypes_by_file_ext=mimetypes_by_file_ext,
        mimetype_detection_hook=mimetype_detection_hook,
        key_from_filename=key_from_filename,
        filter=filter,
    )
    path = Path(path)
    parsed_walkers = []
    for walker in walkers or []:
        parsed_walkers.append(import_object(walker))
    parsed_walkers.extend(DEFAULT_WALKERS)
    prefix_parts = [segment for segment in prefix.split("/") if segment]
    for segment in prefix_parts:
        child_node = await anyio.to_thread.run_sync(node.get, segment)
        if child_node is None:
            key = key_from_filename(segment)
            await create_node_or_drop_collision(
                node,
                structure_family=StructureFamily.container,
                data_sources=[],
                metadata={},
                specs=[],
                key=key,
                access_tags=[],
            )
            # TODO When we have a tiled AsyncClient, use that.
            child_node = await anyio.to_thread.run_sync(node.get, segment)
        node = child_node
    if path.is_dir():
        # Recursively enter the directory and any subdirectories.
        if overwrite:
            logger.info(f"  Overwriting '/{'/'.join(prefix_parts)}'")
            # TODO When we have a tiled AsyncClient, use that.
            await anyio.to_thread.run_sync(partial(node.delete, recursive=True))
        await _walk(
            node,
            Path(path),
            parsed_walkers,
            settings=settings,
        )
    else:
        await register_single_item(
            node,
            path,
            is_directory=False,
            settings=settings,
        )


async def _walk(
    node,
    path,
    walkers,
    settings,
):
    "This is the recursive inner loop of walk."
    files = []
    directories = []
    logger.info("  Walking '%s'", path)
    for item in path.iterdir():
        if not settings.filter(item):
            continue
        if item.is_dir():
            directories.append(item)
        else:
            files.append(item)
    for walker in walkers:
        files, directories = await walker(
            node,
            path,
            files,
            directories,
            settings,
        )
    for directory in directories:
        key = settings.key_from_filename(directory.name)
        await create_node_or_drop_collision(
            node,
            key=key,
            structure_family=StructureFamily.container,
            data_sources=[],
            metadata={},
            specs=[],
            access_tags=[],
        )
        # TODO When we have a tiled AsyncClient, use that.
        child_node = await anyio.to_thread.run_sync(node.get, key)
        await _walk(
            child_node,
            directory,
            walkers,
            settings,
        )


async def one_node_per_item(
    node,
    path,
    files,
    directories,
    settings,
):
    "Process each file and directory as mapping to one logical 'node' in Tiled."
    unhandled_files = []
    unhandled_directories = []
    for file in files:
        result = await register_single_item(
            node,
            file,
            is_directory=False,
            settings=settings,
        )
        if result is None:
            unhandled_files.append(file)
    for directory in directories:
        result = await register_single_item(
            node,
            directory,
            is_directory=True,
            settings=settings,
        )
        if result is None:
            unhandled_directories.append(directory)
    return unhandled_files, unhandled_directories


async def register_single_item(
    node,
    item,
    is_directory,
    settings,
):
    "Register a single file or directory as a node."
    unhandled_items = []
    mimetype = resolve_mimetype(
        item, settings.mimetypes_by_file_ext, settings.mimetype_detection_hook
    )
    if mimetype is None:
        unhandled_items.append(item)
        if not is_directory:
            logger.info("    SKIPPED: Could not resolve mimetype for '%s'", item)
        return
    if mimetype not in settings.adapters_by_mimetype:
        logger.info(
            "    SKIPPED: Resolved mimetype '%s' but no adapter found for '%s'",
            mimetype,
            item,
        )
        unhandled_items.append(item)
        return
    adapter_class = settings.adapters_by_mimetype[mimetype]
    logger.info("    Resolved mimetype '%s' with adapter for '%s'", mimetype, item)
    try:
        adapter = await anyio.to_thread.run_sync(
            adapter_class.from_uris, ensure_uri(item)
        )
    except Exception:
        logger.exception("    SKIPPED: Error constructing adapter for '%s':", item)
        return
    key = settings.key_from_filename(item.name)
    if hasattr(adapter, "generate_data_sources"):
        # Let the Adapter describe the DataSouce(s).
        data_sources = adapter.generate_data_sources(
            mimetype, dict_or_none, item, is_directory
        )
    else:
        # Back-compat: Assume one Asset passed as a
        # parameter named 'data_uri'.
        data_sources = [
            DataSource(
                structure_family=adapter.structure_family,
                mimetype=mimetype,
                structure=dict_or_none(adapter.structure()),
                parameters={},
                management=Management.external,
                assets=[
                    Asset(
                        data_uri=ensure_uri(item),
                        is_directory=is_directory,
                        parameter="data_uri",
                    )
                ],
            )
        ]
    return await create_node_or_drop_collision(
        node,
        key=key,
        structure_family=adapter.structure_family,
        metadata=dict(adapter.metadata()),
        specs=adapter.specs,
        data_sources=data_sources,
    )


# Matches filename with (optional) prefix characters followed by digits \d
# and then the file extension .tif or .tiff.
IMG_SEQUENCE_STEM_PATTERNS = {
    ".tif": re.compile(r"^(.*?)(\d+)\.(?:tif|tiff)$"),
    ".tiff": re.compile(r"^(.*?)(\d+)\.(?:tif|tiff)$"),
    ".jpg": re.compile(r"^(.*?)(\d+)\.(?:jpg|jpeg)$"),
    ".jpeg": re.compile(r"^(.*?)(\d+)\.(?:jpg|jpeg)$"),
    ".npy": re.compile(r"^(.*?)(\d+)\.npy$"),
}
IMG_SEQUENCE_EMPTY_NAME_ROOT = "_unnamed"

IMG_SEQUENCE_MIMETYPES = {
    ".tif": "multipart/related;type=image/tiff",
    ".tiff": "multipart/related;type=image/tiff",
    ".jpg": "multipart/related;type=image/npy",
    ".jpeg": "multipart/related;type=image/jpeg",
    ".npy": "multipart/related;type=application/x-npy",
}


async def group_image_sequences(
    node,
    path,
    files,
    directories,
    settings,
):
    """
    Group files in the given directory into TIFF sequences.

    We are looking for any files:
    - with file extension .tif or .tiff
    - with file name ending in a number

    We group these into sorted groups and make one Node for each.
    A group may have one or more items.
    """
    unhandled_directories = directories
    unhandled_files = []
    sequences = collections.defaultdict(list)
    for file in files:
        file_ext = Path(file).suffixes[-1] if len(Path(file).suffixes) > 0 else None
        if file.is_file() and file_ext and file_ext in IMG_SEQUENCE_STEM_PATTERNS:
            match = IMG_SEQUENCE_STEM_PATTERNS[file_ext].match(file.name)
            if match:
                sequence_name, _sequence_number = match.groups()
                if sequence_name == "":
                    sequence_name = IMG_SEQUENCE_EMPTY_NAME_ROOT
                sequences[sequence_name].append(file)
                continue
        unhandled_files.append(file)
    for name, sequence in sequences.items():
        await register_image_sequence(node, name, sorted(sequence), settings)
    return unhandled_files, unhandled_directories


async def register_image_sequence(node, name, sequence, settings):
    suffixes = Path(sequence[0]).suffixes
    file_ext = suffixes[-1] if len(suffixes) > 0 else None
    if file_ext and file_ext in IMG_SEQUENCE_MIMETYPES:
        mimetype = IMG_SEQUENCE_MIMETYPES[file_ext]
    else:
        raise RuntimeError(f"Cannot register image sequence {name}.")
    logger.info(
        "    Grouped %d %s images into a sequence '%s'",
        len(sequence),
        mimetype.split("/")[-1],
        name,
    )
    adapter_class = settings.adapters_by_mimetype[mimetype]
    key = settings.key_from_filename(name)
    try:
        adapter = adapter_class.from_uris(
            *[ensure_uri(filepath) for filepath in sequence]
        )
    except Exception:
        logger.exception("    SKIPPED: Error constructing adapter for '%s'", name)
        return
    await create_node_or_drop_collision(
        node,
        key=key,
        structure_family=adapter.structure_family,
        metadata=dict(adapter.metadata()),
        specs=adapter.specs,
        data_sources=[
            DataSource(
                structure_family=adapter.structure_family,
                mimetype=mimetype,
                structure=dict_or_none(adapter.structure()),
                parameters={},
                management=Management.external,
                assets=[
                    Asset(
                        data_uri=ensure_uri(item),
                        is_directory=False,
                        parameter="data_uris",
                        num=i,
                    )
                    for i, item in enumerate(sequence)
                ],
            )
        ],
    )


async def skip_all(
    node,
    path,
    files,
    directories,
    settings,
):
    """
    Skip all files and directories without processing them.

    This can be used to override the DEFAULT_WALKERS.
    """
    for item in files:
        logger.info("    SKIP ALL: Nothing yet handled file '%s'", item)
    return [], directories


DEFAULT_WALKERS = [group_image_sequences, one_node_per_item]


async def watch(
    node,
    path,
    prefix="/",
    walkers=None,
    adapters_by_mimetype=None,
    mimetypes_by_file_ext=None,
    mimetype_detection_hook=None,
    key_from_filename=None,
    filter=None,
    initial_walk_complete_event=None,
):
    settings = Settings.init(
        adapters_by_mimetype=adapters_by_mimetype,
        mimetypes_by_file_ext=mimetypes_by_file_ext,
        mimetype_detection_hook=mimetype_detection_hook,
        key_from_filename=key_from_filename,
        filter=filter,
    )
    if initial_walk_complete_event is None:
        initial_walk_complete_event = anyio.Event()
    ready_event = anyio.Event()
    stop_event = anyio.Event()
    async with anyio.create_task_group() as tg:
        # Begin listening for changes.
        tg.start_soon(
            _watch,
            ready_event,
            initial_walk_complete_event,
            stop_event,
            node,
            path,
            prefix,
            walkers,
            settings,
        )
        await ready_event.wait()
        # We have begun listening for changes.
        # Now do the initial walk.
        await register(
            node,
            path,
            prefix,
            walkers,
            adapters_by_mimetype=settings.adapters_by_mimetype,
            mimetypes_by_file_ext=settings.mimetypes_by_file_ext,
            mimetype_detection_hook=settings.mimetype_detection_hook,
            key_from_filename=settings.key_from_filename,
            filter=settings.filter,
        )
        # Signal that initial walk is complete.
        # Process any changes that were accumulated during the initial walk.
        initial_walk_complete_event.set()


async def _watch(
    ready_event,
    initial_walk_complete_event,
    stop_event,
    node,
    path,
    prefix,
    walkers,
    settings,
):
    def watch_filter(change, path):
        return settings.filter(Path(path))

    ready_event.set()
    backlog = []
    async for batch in watchfiles.awatch(
        path,
        watch_filter=watch_filter,
        yield_on_timeout=True,
        stop_event=stop_event,
        rust_timeout=1000,
    ):
        if (backlog is not None) and batch:
            logger.info(
                "Detected changes, waiting to process until initial walk completes"
            )
            backlog.extend(batch)
        if (backlog is not None) and initial_walk_complete_event.is_set():
            logger.info("Watching for changes in '%s'", path)
            if backlog:
                logger.info(
                    "Processing backlog of changes that occurred during initial walk..."
                )
                await process_changes(
                    batch,
                    node,
                    path,
                    prefix,
                    walkers,
                    settings,
                )
            backlog = None
        elif batch:
            # We are caught up. Process changes immediately.
            logger.info("Detected changes")
            await process_changes(
                batch,
                node,
                path,
                prefix,
                walkers,
                settings,
            )


async def process_changes(
    batch,
    node,
    path,
    prefix,
    walkers,
    settings,
):
    for change in batch:
        change_type, change_path = change
        logger.info("  %s '%s'", change_type.name, change_path)
        # TODO Be more selective.
        # We should be able to re-register only a select portion of the
        # full tree. For now, we ignore the change batch content and just
        # use the change as a timing signal to re-register the whole tree.
        await register(
            node,
            path,
            prefix,
            walkers,
            adapters_by_mimetype=settings.adapters_by_mimetype,
            mimetypes_by_file_ext=settings.mimetypes_by_file_ext,
            mimetype_detection_hook=settings.mimetype_detection_hook,
            key_from_filename=settings.key_from_filename,
            filter=settings.filter,
        )


async def create_node_or_drop_collision(
    node,
    *args,
    key,
    **kwargs,
):
    "Call node.new(...) and if there is a collision remove the original."

    # TODO When we have a tiled AsyncClient, use that.
    # run_sync does not accept kwargs, so we make a closure here.
    def new(*args):
        return node.new(*args, key=key, **kwargs)

    try:
        return await anyio.to_thread.run_sync(new, *args)
    except ClientError as err:
        if err.response.status_code == httpx.codes.CONFLICT:
            # To avoid ambiguity include _neither_ the original nor the new one.
            offender = await anyio.to_thread.run_sync(node.get, key)
            await anyio.to_thread.run_sync(partial(offender.delete, recursive=True))
            logger.warning(
                "   COLLISION: Multiple files would result in node at '%s'. Skipping all.",
                err.args[0],
            )
        else:
            raise


def dict_or_none(structure):
    if structure is None:
        return None
    return dataclasses.asdict(structure)



import collections
import collections.abc
import warnings
from urllib.parse import parse_qs, urlparse

import httpx

from ..utils import import_object, prepend_to_sys_path
from .container import DEFAULT_STRUCTURE_CLIENT_DISPATCH, Container
from .context import DEFAULT_TIMEOUT_PARAMS, UNSET, Context
from .utils import MSGPACK_MIME_TYPE, client_for_item, handle_error, retry_context


def from_uri(
    uri,
    structure_clients="numpy",
    *,
    cache=UNSET,
    remember_me=True,
    username=None,
    auth_provider=None,
    api_key=None,
    verify=True,
    prompt_for_reauthentication=None,
    headers=None,
    timeout=None,
    include_data_sources=False,
):
    """
    Connect to a Node on a local or remote server.

    Parameters
    ----------
    uri : str
        e.g. "http://localhost:8000"
    structure_clients : str or dict, optional
        Use "dask" for delayed data loading and "numpy" for immediate
        in-memory structures (e.g. normal numpy arrays, pandas
        DataFrames). For advanced use, provide dict mapping a
        structure_family or a spec to a client object.
    cache : Cache, optional
    remember_me : bool
        Next time, try to automatically authenticate using this session.
    username : str, optional
        DEPRECATED. Ignored, and issues a warning if passed.
    auth_provider : str, optional
        DEPRECATED. Ignored, and issues a warning if passed.
    api_key : str, optional
        API key based authentication.
    verify : bool, optional
        Verify SSL certifications. True by default. False is insecure,
        intended for development and testing only.
    prompt_for_reauthentication : bool, optional
        DEPRECATED. Ignored, and issue a warning if passed.
    headers : dict, optional
        Extra HTTP headers.
    timeout : httpx.Timeout, optional
        If None, use Tiled default settings.
        (To disable timeouts, use httpx.Timeout(None)).
    include_data_sources : bool, optional
        Default False. If True, fetch information about underlying data sources.
    """
    EXPLAIN_LOGIN = """

The user will be prompted for credentials if login is required.
Or, call login() to manually login.

For non-interactive authentication, use an API key.
"""
    if username is not None:
        warnings.warn("Tiled no longer accepts 'username' parameter. " + EXPLAIN_LOGIN)
    if auth_provider is not None:
        warnings.warn(
            "Tiled no longer accepts 'auth_provider' parameter. " + EXPLAIN_LOGIN
        )
    if auth_provider is not None:
        warnings.warn(
            "Tiled no longer accepts 'prompt_for_reauthentication' parameter. "
            + EXPLAIN_LOGIN
        )
    context, node_path_parts = Context.from_any_uri(
        uri,
        api_key=api_key,
        cache=cache,
        headers=headers,
        timeout=timeout,
        verify=verify,
    )
    return from_context(
        context,
        structure_clients=structure_clients,
        node_path_parts=node_path_parts,
        include_data_sources=include_data_sources,
        remember_me=remember_me,
    )


def from_context(
    context: Context,
    structure_clients="numpy",
    node_path_parts=None,
    include_data_sources=False,
    remember_me=True,
):
    """
    Advanced: Connect to a Node using a custom instance of httpx.Client or httpx.AsyncClient.

    Parameters
    ----------
    context : tiled.client.context.Context
    structure_clients : str or dict, optional
        Use "dask" for delayed data loading and "numpy" for immediate
        in-memory structures (e.g. normal numpy arrays, pandas
        DataFrames). For advanced use, provide dict mapping a
        structure_family or a spec to a client object.
    """
    node_path_parts = node_path_parts or []
    # Do entrypoint discovery if it hasn't yet been done.
    if Container.STRUCTURE_CLIENTS_FROM_ENTRYPOINTS is None:
        Container.discover_clients_from_entrypoints()
    # Interpret structure_clients="numpy" and structure_clients="dask" shortcuts.
    if isinstance(structure_clients, str):
        structure_clients = DEFAULT_STRUCTURE_CLIENT_DISPATCH[structure_clients]
    # To construct a user-facing client object, we may be required to authenticate.
    # 1. If any API key set, we are already authenticated and there is nothing to do.
    # 2. If there are cached valid credentials for this server, use them.
    # 3. If not, and the server requires authentication, prompt for authentication.
    if context.api_key is None:
        auth_is_required = context.server_info.authentication.required
        has_providers = len(context.server_info.authentication.providers) > 0
        if auth_is_required and not has_providers:
            raise RuntimeError(
                """This server requires API key authentication.
    Set an api_key as in:

    >>> c = from_uri("...", api_key="...")
    """
            )
        if has_providers:
            found_valid_tokens = remember_me and context.use_cached_tokens()
            if (not found_valid_tokens) and auth_is_required:
                context.authenticate(remember_me=remember_me)
    # Context ensures that context.api_uri has a trailing slash.
    item_uri = f"{context.api_uri}metadata/{'/'.join(node_path_parts)}"
    params = parse_qs(urlparse(item_uri).query)
    if include_data_sources:
        params["include_data_sources"] = include_data_sources
    for attempt in retry_context():
        with attempt:
            content = handle_error(
                context.http_client.get(
                    item_uri,
                    headers={"Accept": MSGPACK_MIME_TYPE},
                )
            ).json()
    item = content["data"]
    return client_for_item(
        context, structure_clients, item, include_data_sources=include_data_sources
    )


def from_profile(name, structure_clients=None, **kwargs):
    """
    Build a Node based a 'profile' (a named configuration).

    List available profiles and the source filepaths from Python like:

    >>> from tiled.profiles import list_profiles
    >>> list_profiles()

    or from a CLI like:

    $ tiled profile list

    Or show the file contents like:

    >>> from tiled.profiles import load_profiles
    >>> load_profiles()

    or from a CLI like:

    $ tiled profile show PROFILE_NAME

    Any additional parameters override profile content. See from_uri for details.
    """
    # We accept structure_clients as a separate parameter so that it
    # may be invoked positionally, as in from_profile("...", "dask").
    from ..profiles import ProfileNotFound, load_profiles, paths

    profiles = load_profiles()
    try:
        filepath, profile_content = profiles[name]
    except KeyError as err:
        raise ProfileNotFound(
            f"Profile {name!r} not found. Found profiles {list(profiles)} "
            f"from directories {paths}."
        ) from err
    merged = {**profile_content, **kwargs}
    if structure_clients is not None:
        merged["structure_clients"] = structure_clients
    cache_config = merged.pop("cache", None)
    if cache_config is not None:
        from tiled.client.cache import Cache

        if isinstance(cache_config, collections.abc.Mapping):
            # All necessary validation has already been performed
            # in load_profiles().
            cache = Cache(**cache_config)
        else:
            # Interpret this as a Cache object passed in directly.
            cache = cache_config
        merged["cache"] = cache
    timeout_config = merged.pop("timeout", None)
    if timeout_config is not None:
        if isinstance(timeout_config, httpx.Timeout):
            timeout = timeout_config
        else:
            timeout_params = DEFAULT_TIMEOUT_PARAMS.copy()
            timeout_params.update(timeout_config)
            timeout = httpx.Timeout(**timeout_params)
        merged["timeout"] = timeout
    # Below, we may convert importable strings like
    # "package.module:obj" to live objects. Include the profile's
    # source directory in the import path, temporarily.
    with prepend_to_sys_path(filepath.parent):
        structure_clients_ = merged.pop("structure_clients", None)
        if structure_clients_ is not None:
            if isinstance(structure_clients_, str):
                # Nothing to do.
                merged["structure_clients"] = structure_clients_
            else:
                # This is a dict mapping structure families like "array" and "dataframe"
                # to values. The values may be client objects or importable strings.
                result = {}
                for key, value in structure_clients_.items():
                    if isinstance(value, str):
                        class_ = import_object(value, accept_live_object=True)
                    else:
                        class_ = value
                    result[key] = class_
                merged["structure_clients"] = result
    if "direct" in merged:
        # The profile specifies the server in-line.
        # Create an app and use it directly via ASGI.
        import jsonschema

        from ..config import ConfigError, schema
        from ..server.app import build_app_from_config

        config = merged.pop("direct", None)
        try:
            jsonschema.validate(instance=config, schema=schema())
        except jsonschema.ValidationError as err:
            msg = err.args[0]
            raise ConfigError(
                f"ValidationError while parsing configuration file {filepath}: {msg}"
            ) from err
        context = Context.from_app(
            build_app_from_config(config, source_filepath=filepath),
        )
        return from_context(context, **merged)
    else:
        return from_uri(**merged)



import builtins
from typing import Any, List, Optional, Set, Tuple, Union

import numpy as np
from numpy._typing import NDArray
from PIL import Image

from ..catalog.orm import Node
from ..ndslice import NDSlice
from ..storage import Storage
from ..structures.array import ArrayStructure, BuiltinDtype
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import DataSource
from ..type_aliases import JSON
from ..utils import path_from_uri
from .resource_cache import with_resource_cache
from .sequence import FileSequenceAdapter
from .utils import init_adapter_from_catalog


class JPEGAdapter:
    """
    Read a JPEG file.

    Examples
    --------

    >>> JPEGAdapter("path/to/file.jpeg")
    """

    structure_family = StructureFamily.array
    supported_storage: Set[type[Storage]] = set()

    def __init__(
        self,
        data_uri: str,
        *,
        structure: ArrayStructure,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        data_uri :
        structure :
        metadata :
        specs :
        """
        filepath = path_from_uri(data_uri)
        cache_key = (Image.open, filepath)
        self._file = with_resource_cache(cache_key, Image.open, filepath)
        self.specs = specs or []
        self._provided_metadata = metadata or {}
        self._structure = structure

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[ArrayStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "JPEGAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    @classmethod
    def from_uris(
        cls,
        data_uri: str,
        **kwargs: Optional[Any],
    ) -> "JPEGAdapter":
        filepath = path_from_uri(data_uri)
        cache_key = (Image.open, filepath)
        _file = with_resource_cache(cache_key, Image.open, filepath)

        arr = np.asarray(_file)
        structure = ArrayStructure(
            shape=arr.shape,
            chunks=tuple((dim,) for dim in arr.shape),
            data_type=BuiltinDtype.from_numpy_dtype(arr.dtype),
        )

        return cls(
            data_uri,
            structure=structure,
        )

    def metadata(self) -> JSON:
        return self._provided_metadata.copy()

    def read(self, slice: NDSlice = NDSlice(...)) -> NDArray[Any]:
        arr = np.asarray(self._file)
        arr = arr[slice] if slice else arr
        return arr

    def read_block(
        self, block: Tuple[int, ...], slice: Optional[builtins.slice] = None
    ) -> NDArray[Any]:
        if sum(block) != 0:
            raise IndexError(block)

        arr = np.asarray(self._file)
        if slice is not None:
            arr = arr[slice]
        return arr

    def structure(self) -> ArrayStructure:
        return self._structure


class JPEGSequenceAdapter(FileSequenceAdapter):
    def _load_from_files(
        self, slice: Union[builtins.slice, int] = slice(None)
    ) -> NDArray[Any]:
        from PIL import Image

        if isinstance(slice, int):
            return np.asarray(Image.open(self.filepaths[slice]))[None, ...]
        else:
            return np.asarray(
                [np.asarray(Image.open(file)) for file in self.filepaths[slice]]
            )



"""
A directory containing awkward buffers, one file per form key.
"""

import copy
from pathlib import Path
from typing import Any, List, Optional
from urllib.parse import quote_plus

import awkward.forms

from ..catalog.orm import Node
from ..storage import FileStorage, Storage
from ..structures.awkward import AwkwardStructure
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import Asset, DataSource
from ..type_aliases import JSON
from ..utils import path_from_uri
from .awkward import AwkwardAdapter
from .awkward_directory_container import DirectoryContainer
from .utils import init_adapter_from_catalog


class AwkwardBuffersAdapter(AwkwardAdapter):
    structure_family = StructureFamily.awkward
    supported_storage = {FileStorage}

    @classmethod
    def init_storage(
        cls,
        storage: Storage,
        data_source: DataSource[AwkwardStructure],
        path_parts: List[str],
    ) -> DataSource[AwkwardStructure]:
        """

        Parameters
        ----------
        data_uri :
        structure :

        Returns
        -------

        """
        data_source = copy.deepcopy(data_source)  # Do not mutate caller input.
        data_uri = storage.uri + "".join(
            f"/{quote_plus(segment)}" for segment in path_parts
        )
        directory: Path = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)
        data_source.assets.append(
            Asset(data_uri=data_uri, is_directory=True, parameter="data_uri")
        )
        return data_source

    def __init__(
        self,
        data_uri: str,
        structure: AwkwardStructure,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ):
        form = awkward.forms.from_dict(structure.form)
        directory: Path = path_from_uri(data_uri)
        if not directory.is_dir():
            raise ValueError(f"Not a directory: {directory}")
        container = DirectoryContainer(directory, form)
        super().__init__(
            container,
            structure=structure,
            metadata=metadata,
            specs=specs,
        )

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[AwkwardStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "AwkwardBuffersAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore



"""Create sqlite table for fulltext search

Revision ID: ed3a4223a600
Revises: e756b9381c14
Create Date: 2024-04-11 16:41:01.369520

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ed3a4223a600"
down_revision = "e756b9381c14"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()
    if connection.engine.dialect.name == "sqlite":
        statements = [
            # Create an external content fts5 table.
            # See https://www.sqlite.org/fts5.html Section 4.4.3.
            """
            CREATE VIRTUAL TABLE metadata_fts5 USING fts5(metadata, content='nodes', content_rowid='id');
            """,
            # Insert all existing node content into the fts5 table.
            """
            INSERT INTO metadata_fts5(rowid, metadata)
            SELECT id, metadata FROM nodes;
            """,
            # Triggers keep the index synchronized with the nodes table.
            """
            CREATE TRIGGER nodes_metadata_fts5_sync_ai AFTER INSERT ON nodes BEGIN
              INSERT INTO metadata_fts5(rowid, metadata)
              VALUES (new.id, new.metadata);
            END;
            """,
            """
            CREATE TRIGGER nodes_metadata_fts5_sync_ad AFTER DELETE ON nodes BEGIN
              INSERT INTO metadata_fts5(metadata_fts5, rowid, metadata)
              VALUES('delete', old.id, old.metadata);
            END;
            """,
            """
            CREATE TRIGGER nodes_metadata_fts5_sync_au AFTER UPDATE ON nodes BEGIN
              INSERT INTO metadata_fts5(metadata_fts5, rowid, metadata)
              VALUES('delete', old.id, old.metadata);
              INSERT INTO metadata_fts5(rowid, metadata)
              VALUES (new.id, new.metadata);
            END;
            """,
        ]
        for statement in statements:
            op.execute(sa.text(statement))


def downgrade():
    pass



"""Demote `Composite` structure_familty to `composite spec`

Revision ID: dfbb7478c6bd
Revises: a963a6c32a0c
Create Date: 2025-08-18 17:56:14.996646

"""
import json
from dataclasses import asdict

import sqlalchemy as sa
from alembic import op

from tiled.structures.core import Spec

# revision identifiers, used by Alembic.
revision = "dfbb7478c6bd"
down_revision = "a963a6c32a0c"
branch_labels = None
depends_on = None

# Representations of Spec objects as JSON strings
json_spec_composite = json.dumps(asdict(Spec("composite")))
json_spec_flattened = json.dumps(asdict(Spec("flattened")))


def upgrade():
    connection = op.get_bind()

    # Step 1: Update the node wherever structure_family = 'composite':
    # - Add 'composite' spec to specs, in the end
    # - Change structure_family to 'container'
    if connection.engine.dialect.name == "postgresql":
        json_stmnt = f"specs || '[{json_spec_composite}]'::jsonb"
    elif connection.engine.dialect.name == "sqlite":
        json_stmnt = f"json_insert(specs, '$[#]', json('{json_spec_composite}'))"
    op.execute(
        f"""
        UPDATE nodes SET specs = {json_stmnt},
            structure_family = 'container'
        WHERE structure_family = 'composite'
        """
    )

    # Step 2. Update the data_sources table (usually unnecessary)
    op.execute(
        """
        UPDATE data_sources
        SET structure_family = 'container'
        WHERE structure_family = 'composite'
        """
    )

    # Step 3. Remove 'flattened' spec used for tables in Composite nodes
    if connection.engine.dialect.name == "postgresql":
        op.execute(
            f"""
            UPDATE nodes
            SET specs = (
                SELECT COALESCE(jsonb_agg(elem), '[]'::jsonb)
                FROM jsonb_array_elements(specs) AS arr(elem)
                WHERE elem != '{json_spec_flattened}'::jsonb
            )
            WHERE specs @> '[{json_spec_flattened}]'::jsonb
            """
        )
    elif connection.engine.dialect.name == "sqlite":
        op.execute(
            f"""
            UPDATE nodes
            SET specs = (
                SELECT json_group_array(value)
                FROM json_each(specs)
                WHERE value != json('{json_spec_flattened}')
            )
            WHERE EXISTS (
                SELECT 1
                FROM json_each(specs)
                WHERE value = json('{json_spec_flattened}')
            )
            """
        )

    # Step 4. Remove 'composite' structure family from the `structurefamily` enum
    if connection.engine.dialect.name == "postgresql":
        # 4.1. Rename the existing type
        op.execute("ALTER TYPE structurefamily RENAME TO structurefamily_old")

        # 4.2. Get all enum values from the existing type
        values = connection.execute(
            sa.text("SELECT unnest(enum_range(NULL::structurefamily_old))")
        ).fetchall()
        values = [r[0] for r in values if r[0] != "composite"]

        # 4.3. Create the new type (without 'composite')
        op.execute(
            sa.text(
                "CREATE TYPE structurefamily AS ENUM ("
                + ", ".join(f"'{v}'" for v in values)
                + ")"
            )
        )

        # 4.4. Alter columns to use the new type
        for table_name in ["nodes", "data_sources"]:
            op.execute(
                f"""
                ALTER TABLE {table_name}
                ALTER COLUMN structure_family TYPE structurefamily
                USING structure_family::text::structurefamily
                """
            )

        # 4.5. Drop the old type
        op.execute("DROP TYPE structurefamily_old")


def downgrade():
    connection = op.get_bind()

    # Step 1: Re-add 'composite' to the enum
    if connection.engine.dialect.name == "postgresql":
        with op.get_context().autocommit_block():
            op.execute(
                """
            ALTER TYPE structurefamily
            ADD VALUE IF NOT EXISTS 'composite' AFTER 'awkward'
            """
            )

    # Step 2: Revert data changes
    # Look for rows whose specs include the 'composite' spec
    json_spec_composite = json.dumps(asdict(Spec("composite")))

    if connection.engine.dialect.name == "postgresql":
        # Revert structure_family
        op.execute(
            f"""
            UPDATE nodes SET structure_family = 'composite'
            WHERE specs @> '[{json_spec_composite}]'::jsonb
            """
        )
        # Remove the 'composite' entry from specs
        op.execute(
            f"""
            UPDATE nodes
            SET specs = (
                SELECT COALESCE(jsonb_agg(elem), '[]'::jsonb)
                FROM jsonb_array_elements(specs) AS arr(elem)
                WHERE elem != '{json_spec_composite}'::jsonb
            )
            WHERE specs @> '[{json_spec_composite}]'::jsonb
            """
        )

    elif connection.engine.dialect.name == "sqlite":
        # Revert structure_family
        op.execute(
            f"""
            UPDATE nodes
            SET structure_family = 'composite'
            WHERE EXISTS (
                SELECT 1 FROM json_each(specs)
                WHERE value = json('{json_spec_composite}')
            )
            """
        )
        # Remove the 'composite' entry from specs
        op.execute(
            f"""
            UPDATE nodes
            SET specs = (
                SELECT json_group_array(value)
                FROM json_each(specs)
                WHERE value != json('{json_spec_composite}')
            )
            WHERE EXISTS (
                SELECT 1
                FROM json_each(specs)
                WHERE value = json('{json_spec_composite}')
            )
            """
        )

    # Step 3: Mark tables belonging to composite nodes as 'flattened'
    if connection.engine.dialect.name == "postgresql":
        json_stmnt = f"specs || '[{json_spec_flattened}]'::jsonb"
    elif connection.engine.dialect.name == "sqlite":
        json_stmnt = f"json_insert(specs, '$[#]', json('{json_spec_flattened}'))"
    op.execute(
        f"""
        UPDATE nodes SET specs = {json_stmnt}
        WHERE structure_family = 'table'
            AND EXISTS (
                SELECT 1
                FROM nodes AS parent
                WHERE parent.id = nodes.parent
                    AND parent.structure_family = 'composite'
            )
        """
    )

    # Step 4. Update the data_sources table (usually unnecessary)
    op.execute(
        """
        UPDATE data_sources
        SET structure_family = 'composite'
        WHERE structure_family = 'container'
            AND EXISTS (
                SELECT 1
                FROM nodes
                WHERE nodes.id = data_sources.node_id
                    AND nodes.structure_family = 'composite'
            )
        """
    )



import hashlib
from pathlib import Path

import pandas
import pytest
from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
)

from ..catalog import in_memory
from ..client import Context, from_context
from ..client.utils import get_asset_filepaths
from ..server.app import build_app
from ..utils import path_from_uri
from .utils import fail_with_status_code


@pytest.fixture
def context(tmpdir):
    catalog = in_memory(writable_storage=str(tmpdir / "data"))
    app = build_app(catalog)
    with Context.from_app(app) as context:
        yield context


@pytest.fixture
def client(context):
    client = from_context(context)
    yield client


def test_include_data_sources_method_on_self(client):
    "Calling include_data_sources() fetches data sources on self."
    x = client.write_array([1, 2, 3], key="x")
    # Fetch data_sources on x object directly.
    assert x.data_sources() is not None
    # Fetch data_sources on x object, looked up in client.
    assert client["x"].data_sources() is not None
    assert client["x"].include_data_sources().data_sources() is not None


def test_include_data_sources_method_affects_children(client):
    "Calling include_data_sources() fetches data sources on children."
    client.create_container("c")
    client["c"].write_array([1, 2, 3], key="x")
    c = client["c"].include_data_sources()
    assert c["x"].data_sources() is not None


def test_include_data_sources_kwarg(context):
    "Passing include_data_sources to constructor includes them by default."
    client = from_context(context, include_data_sources=True)
    client.write_array([1, 2, 3], key="x")
    assert client["x"].data_sources() is not None
    assert client["x"].include_data_sources() is not None


def test_raw_export(client, tmpdir):
    "Use raw_export() and compare hashes or original and exported files."
    client.write_array([1, 2, 3], key="x")
    exported_paths = client["x"].raw_export(tmpdir / "exported")
    data_sources = client["x"].include_data_sources().data_sources()
    orig_dir = path_from_uri(data_sources[0].assets[0].data_uri)
    _asset_id, relative_paths = client["x"].asset_manifest(data_sources).popitem()
    orig_paths = [Path(orig_dir, relative_path) for relative_path in relative_paths]
    orig_hashes = [hashlib.md5(path.read_bytes()).digest() for path in orig_paths]
    exported_hashes = [
        hashlib.md5(path.read_bytes()).digest() for path in exported_paths
    ]
    assert orig_hashes == exported_hashes


def test_asset_range_request(client, tmpdir):
    "Access part of an asset using an HTTP Range header."
    df = pandas.DataFrame({"A": [1, 2, 3], "B": [4.0, 5.0, 6.0]})
    client.write_dataframe(df, key="x")
    # Fetch the first byte.
    first_byte_response = client.context.http_client.get(
        "/api/v1/asset/bytes/x?id=1",
        headers={"Range": "bytes=0-0"},
    )
    assert first_byte_response.content == b"P"
    # Fetch the first two bytes.
    first_two_bytes_response = client.context.http_client.get(
        "/api/v1/asset/bytes/x?id=1",
        headers={"Range": "bytes=0-1"},
    )
    assert first_two_bytes_response.content == b"PA"
    # Fetch the second two bytes.
    second_two_bytes_response = client.context.http_client.get(
        "/api/v1/asset/bytes/x?id=1",
        headers={"Range": "bytes=2-3"},
    )
    assert second_two_bytes_response.content == b"R1"
    # Request outside of range
    out_of_range_response = client.context.http_client.get(
        "/api/v1/asset/bytes/x?id=1",
        headers={"Range": "bytes=1000000-100000000"},
    )
    with fail_with_status_code(HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE):
        out_of_range_response.raise_for_status()
    # Request malformed range
    malformed_response = client.context.http_client.get(
        "/api/v1/asset/bytes/x?id=1",
        headers={"Range": "bytes=abc"},
    )
    with fail_with_status_code(HTTP_400_BAD_REQUEST):
        malformed_response.raise_for_status()


def test_get_asset_filepaths(client):
    "Smoke test get_asset_filepaths."
    client.write_array([1, 2, 3], key="x")
    get_asset_filepaths(client.include_data_sources()["x"])


def test_do_not_expose_raw_assets(tmpdir):
    catalog = in_memory(writable_storage=str(tmpdir / "data"))
    app = build_app(catalog, server_settings={"expose_raw_assets": False})
    with Context.from_app(app) as context:
        client = from_context(context, include_data_sources=True)
        client.write_array([1, 2, 3], key="x")
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            client["x"].raw_export(tmpdir / "exported")
